from serieux.features.proxy import CommentProxy, LazyProxy


def test_lazy_proxy():
    lazy_value = LazyProxy(lambda: 42)

    assert lazy_value
    assert lazy_value == 42
    assert str(lazy_value) == "42"
    assert repr(lazy_value) == "42"


def test_lazy_arithmetic():
    lazy_a = LazyProxy(lambda: 10)
    lazy_b = LazyProxy(lambda: 5)
    lazy_c = LazyProxy(lambda: -3)

    assert lazy_a + lazy_b == 15
    assert lazy_a - lazy_b == 5
    assert lazy_a * lazy_b == 50
    assert lazy_a / lazy_b == 2.0
    assert lazy_a // lazy_b == 2
    assert lazy_a % lazy_b == 0
    assert lazy_a**lazy_b == 100000
    assert abs(lazy_c) == 3
    assert -lazy_c == 3
    assert +lazy_c == -3


def test_lazy_comparisons():
    lazy_value = LazyProxy(lambda: 42)
    assert lazy_value == 42
    assert lazy_value != 43
    assert lazy_value < 100
    assert lazy_value <= 42
    assert lazy_value > 0
    assert lazy_value >= 42


def test_lazy_list():
    lazy_list = LazyProxy(lambda: [1, 2, 3])

    assert len(lazy_list) == 3
    assert lazy_list[0] == 1
    assert list(lazy_list) == [1, 2, 3]
    assert 2 in lazy_list


def test_commented_proxy():
    obj = [1, 2, 3]
    comment = "This is a comment"
    proxy = CommentProxy(obj, comment)

    # The proxy should behave like the original object
    assert proxy[0] == 1
    assert list(proxy) == [1, 2, 3]
    assert 2 in proxy
    assert str(proxy) == str(obj)
    assert repr(proxy) == repr(obj)

    # The comment should be accessible via the "_" attribute
    assert proxy._ == comment

    # The underlying object should be accessible via _obj
    assert proxy._obj is obj



from dataclasses import dataclass
from typing import Annotated

import pytest

from serieux import deserialize, serialize
from serieux.auto import Auto, Call, MeldedCall
from serieux.exc import SchemaError, ValidationError
from serieux.features.tagset import TaggedUnion, tag_field
from serieux.model import constructed_type

from .definitions import Point


class Funky:
    def __init__(self, x: int, y: bool):
        self.marks = funky(x, y)


def funky(x: int, y: bool) -> str:
    return ("!" if y else ".") * x


def test_auto():
    funk = deserialize(Auto[Funky], {"x": 3, "y": True})
    assert funk.marks == "!!!"


def test_auto_from_init():
    funk = deserialize(Auto[Funky], {"x": 3, "y": True})
    assert funk.marks == "!!!"


def test_auto_callable():
    funk = deserialize(Auto[funky], {"x": 3, "y": True})
    assert funk() == "!!!"


def test_call_callable():
    funk = deserialize(Call[funky], {"x": 3, "y": True})
    assert funk == "!!!"


def test_call_on_type():
    with pytest.raises(TypeError, match=r"Call\[...\] should only wrap callables"):
        deserialize(Call[Funky], {"x": 3, "y": True})


def test_auto_not_serializable():
    with pytest.raises(SchemaError, match="does not specify how to serialize"):
        serialize(Auto[Funky], Funky(x=3, y=True))


def test_auto_no_interference():
    pt = serialize(Auto[Point], Point(1, 2))
    assert pt == {"x": 1, "y": 2}


def test_auto_func_lacks_annotations():
    def f(x, y: int) -> int:
        return x + y

    with pytest.raises(TypeError, match="lacks a type annotation"):
        deserialize(Auto[f], {"x": 1, "y": 2})


def add_them(x: int, y: int):
    return x + y


def mul_them(x: int, y: int) -> int:
    return x * y


def test_tagged_union():
    tu = TaggedUnion[Call[add_them], Call[mul_them]]

    result = deserialize(tu, {tag_field: "add_them", "x": 3, "y": 4})
    assert result == 7

    result = deserialize(tu, {tag_field: "mul_them", "x": 3, "y": 4})
    assert result == 12

    with pytest.raises(ValidationError, match="does not match expected tag"):
        deserialize(tu, {tag_field: "div_them", "x": 3, "y": 4})


def test_constructed_type():
    assert constructed_type(Call[mul_them]) is int
    with pytest.raises(TypeError, match="does not have a return type annotation"):
        constructed_type(Call[add_them])


@dataclass
class Person:
    name: str
    ha: str

    def laugh(self, n: int) -> str:
        return self.ha * n


def test_melded_call():
    cp = MeldedCall(Person, Person.laugh)
    assert cp(name="Alice", n=4, ha="ho") == "hohohoho"


def test_auto_method():
    hafn = deserialize(Auto[Person.laugh], {"name": "Bob", "ha": "hi", "n": 3})
    assert hafn() == "hihihi"


def test_auto_method_no_embed_self():
    hafn = deserialize(
        Annotated[Person.laugh, Call(embed_self=False)],
        {"self": {"name": "Bob", "ha": "hu"}, "n": 3},
    )
    assert hafn == "huhuhu"



from typing import Any

import pytest

from serieux import deserialize, schema, serialize
from serieux.exc import ValidationError
from serieux.features.registered import AutoRegistered, Referenced, Registry, auto_singleton
from tests.definitions import Job

teacher = Job(title="Teacher", yearly_pay=50000)
engineer = Job(title="Engineer", yearly_pay=90000)
artist = Job(title="Artist", yearly_pay=40000)


def test_registry():
    reg = Registry(
        {
            "teacher": teacher,
            "artist": artist,
        }
    )
    assert deserialize(Job @ reg, "teacher") is teacher
    assert serialize(Job @ reg, artist) == "artist"

    with pytest.raises(ValidationError):
        deserialize(Job @ reg, "engineer")

    with pytest.raises(ValidationError):
        serialize(Job @ reg, engineer)


def test_referenced():
    assert deserialize(Referenced[Job], "tests.features.test_registered:teacher") is teacher
    with pytest.raises(ValidationError):
        assert serialize(Referenced[Job], teacher)


def test_referenced_function():
    sym = "tests.features.test_registered:test_referenced_function"
    assert deserialize(Referenced[Any], sym) is test_referenced_function
    assert serialize(Referenced[Any], test_referenced_function) == sym


class Person(AutoRegistered):
    def __init__(self, name, age):
        super().__init__(name)
        self.name = name
        self.age = age


class SuperPerson(Person):
    def __init__(self, name, age, power):
        super().__init__(name, age)
        self.power = power


anita = Person("anita", 76)
bernard = Person("bernard", 73)
charlotte = SuperPerson("charlotte", 33, "spits fire")


def test_registered_serialize():
    assert serialize(Person, anita) == "anita"
    assert serialize(Person, bernard) == "bernard"
    assert serialize(Person, charlotte) == "charlotte"


def test_registered_deserialize():
    assert deserialize(Person, "anita") is anita
    assert deserialize(Person, "bernard") is bernard
    assert deserialize(Person, "charlotte") is charlotte


def test_registered_schema():
    sch = schema(Person).compile(root=False)

    assert sch == {"type": "string", "enum": ["anita", "bernard", "charlotte"]}


class Tool(AutoRegistered):
    pass


@auto_singleton("HAMMER")
class Hammer(Tool):
    def use(self):
        return "bang!"


@auto_singleton
class Saw(Tool):
    def use(self):
        return "zing!"


def test_singleton_deserialize():
    assert deserialize(Tool, "HAMMER") is Hammer
    assert deserialize(Tool, "saw") is Saw
    with pytest.raises(ValidationError):
        deserialize(Tool, "nail")


def test_singleton_serialize():
    assert serialize(Tool, Hammer) == "HAMMER"
    assert serialize(Tool, Saw) == "saw"
    with pytest.raises(ValidationError):
        serialize(Tool, object())


def test_illegal_singleton():
    with pytest.raises(
        TypeError, match="must be a subclass of Registered, but not a direct subclass"
    ):

        @auto_singleton
        class Bloop(AutoRegistered):
            pass



from typing import Literal

import pytest
from ovld import Medley

from serieux import Serieux, dump, load, serialize
from serieux.ctx import Context, Trail
from serieux.exc import ValidationError

from .common import has_312_features, one_test_per_assert
from .definitions import Color, DIDHolder, DotDict, File, Level, LTHolder, Point


@one_test_per_assert
def test_serialize_scalars():
    assert serialize(0) == 0
    assert serialize(12) == 12
    assert serialize(-3.25) == -3.25
    assert serialize("flagada") == "flagada"
    assert serialize(True) is True
    assert serialize(False) is False
    assert serialize(None) is None


@one_test_per_assert
def test_serialize_scalars_conversion():
    assert serialize(float, 10) == 10.0


def test_serialize_point():
    pt = Point(1, 2)
    assert serialize(Point, pt) == {"x": 1, "y": 2}


def test_serialize_list_of_points():
    pts = [Point(1, 2), Point(3, 4)]
    assert serialize(list[Point], pts) == [
        {"x": 1, "y": 2},
        {"x": 3, "y": 4},
    ]


def test_serialize_dict_of_points():
    pts = {"p1": Point(1, 2), "p2": Point(3, 4)}
    assert serialize(dict[str, Point], pts) == {
        "p1": {"x": 1, "y": 2},
        "p2": {"x": 3, "y": 4},
    }


@has_312_features
def test_serialize_tree():
    from .definitions_py312 import Tree

    tree = Tree(
        left=Tree(
            left=1,
            right=Tree(left=Tree(left=2, right=3), right=Tree(left=4, right=5)),
        ),
        right=Tree(left=Tree(left=6, right=7), right=8),
    )
    assert serialize(Tree[int], tree) == {
        "left": {
            "left": 1,
            "right": {
                "left": {"left": 2, "right": 3},
                "right": {"left": 4, "right": 5},
            },
        },
        "right": {"left": {"left": 6, "right": 7}, "right": 8},
    }


def test_serialize_set():
    nums = {1, 2, 3, 4}
    assert serialize(set[int], nums) == [1, 2, 3, 4]


def test_serialize_frozenset():
    nums = frozenset({1, 2, 3, 4})
    assert serialize(frozenset[int], nums) == [1, 2, 3, 4]


def test_serialize_dict_subclass():
    data = DotDict(apple=3, banana=7)
    assert serialize(DotDict[str, int], data) == {"apple": 3, "banana": 7}


class Special(Medley):
    def serialize(self, typ: type[int], value: int, ctx: Context):
        return value * 10

    def serialize(self, typ: type[int], value: str, ctx: Context):
        return value * 2


def test_override():
    ss = (Serieux + Special)()
    assert ss.serialize(int, 3) == 30
    assert ss.serialize(int, "quack") == "quackquack"
    assert ss.serialize(list[int], [1, 2, 3]) == [10, 20, 30]
    assert ss.serialize(list[int], [1, "2", 3]) == [10, "22", 30]
    assert ss.serialize(Point, Point(8, 9)) == {"x": 80, "y": 90}
    assert ss.serialize(3) == 30


class quirkint(int):
    pass


class Quirky(Medley):
    def serialize(self, typ: type[int], value: quirkint, ctx: Context):
        return value * 10


def test_override_quirkint():
    ss = (Serieux + Quirky)()
    assert ss.serialize(int, 3) == 3
    assert ss.serialize(int, quirkint(3)) == 30
    assert ss.serialize(Point, Point(8, 9)) == {"x": 8, "y": 9}
    assert ss.serialize(Point, Point(quirkint(8), 9)) == {"x": 80, "y": 9}


class ExtraWeight(Context):
    weight: int


class WeightedImpl(Medley):
    def serialize(self, typ: type[int], value: int, ctx: ExtraWeight):
        return value + ctx.weight


def test_override_state():
    ss = (Serieux + WeightedImpl)()
    assert ss.serialize(int, 3) == 3
    assert ss.serialize(int, 3, ExtraWeight(10)) == 13
    assert ss.serialize(Point, Point(7, 8)) == {"x": 7, "y": 8}
    assert ss.serialize(Point, Point(7, 8), ExtraWeight(10)) == {"x": 17, "y": 18}


def test_serialize_enum():
    assert serialize(Color, Color.RED) == "red"
    assert serialize(list[Color], [Color.GREEN, Color.BLUE, Color.GREEN]) == [
        "green",
        "blue",
        "green",
    ]


def test_serialize_enum_int():
    assert serialize(Level, Level.MED) == 1
    assert serialize(list[Level], [Level.HI, Level.LO, Level.HI]) == [2, 0, 2]


def test_serialize_literal_enum():
    assert serialize(Literal["red", "green", "blue"], "red") == "red"
    with pytest.raises(
        ValidationError,
        match=r"'yellow' is not a valid option for typing.Literal\['red', 'green', 'blue'\]",
    ):
        serialize(Literal["red", "green", "blue"], "yellow")


def test_serialize_literal_int_enum():
    assert serialize(Literal[1, 2, 7], 2) == 2
    with pytest.raises(
        ValidationError,
        match=r"'3' is not a valid option for typing.Literal\[1, 2, 7\]",
    ):
        serialize(Literal[1, 2, 7], 3)


def test_serialize_literal_mixed_enum():
    lit = Literal[1, True, "quack"]
    assert serialize(lit, 1) == 1
    assert serialize(lit, True) is True
    assert serialize(lit, "quack") == "quack"
    with pytest.raises(ValidationError):
        serialize(lit, 3)
    with pytest.raises(ValidationError):
        serialize(lit, "boop")
    with pytest.raises(ValidationError):
        serialize(lit, False)


def test_serialize_ignore(datapath):
    pth = datapath / "job.yaml"
    f = File(pth)
    assert serialize(File, f) == {"path": str(pth)}


###############
# Error tests #
###############


def test_error_basic():
    with pytest.raises(
        ValidationError, match=r"Cannot serialize object of type 'str' into expected type 'int'"
    ):
        serialize(int, "oh no")


def test_error_dataclass():
    with pytest.raises(
        ValidationError, match=r"Cannot serialize object of type 'str' into expected type 'int'"
    ):
        serialize(Point, Point(x=1, y="oops"), Trail())


@has_312_features
def test_error_serialize_tree():
    from .definitions_py312 import Tree

    tree = Tree(Tree("a", 2), "b")

    with pytest.raises(ValidationError, match=r"At path \.left\.right"):
        serialize(Tree[str], tree, Trail())


def test_error_serialize_list():
    li = [0, 1, 2, 3, "oops", 5, 6]

    with pytest.raises(ValidationError, match=r"At path .4"):
        serialize(list[int], li, Trail())


def test_error_serialize_list_of_lists():
    li = [[0, 1], [2, 3, "oops", 5, 6]]

    with pytest.raises(ValidationError, match=r"At path .1.2"):
        serialize(list[list[int]], li, Trail())


def test_dump_no_dest():
    pt = Point(1, 2)
    assert serialize(Point, pt) == dump(Point, pt)


def test_dump(tmp_path):
    dest = tmp_path / "point.yaml"
    pt = Point(1, 2)
    dump(Point, pt, dest=dest)
    assert load(Point, dest) == pt


def test_dump_str_with_format():
    pt = Point(1, 2)
    s = dump(Point, pt, format="yaml")
    assert s == "x: 1\ny: 2\n"


def test_serialize_recursive_type():
    data = DIDHolder(
        did={
            "a": 1,
            "b": {
                "c": 2,
                "d": 3,
            },
        }
    )
    ser = serialize(DIDHolder, data)
    assert ser["did"] == data.did


def test_serialize_recursive_type_2():
    def p(x):
        return Point(x, x)

    data = LTHolder(lt=[p(1), [p(2), [p(3)], p(4)]])
    ser = serialize(LTHolder, data)
    assert ser["lt"] == [
        {"x": 1, "y": 1},
        [{"x": 2, "y": 2}, [{"x": 3, "y": 3}], {"x": 4, "y": 4}],
    ]



import base64
import json
from functools import cache
from hashlib import sha256
from typing import TYPE_CHECKING, Annotated, Any, Callable, TypeAlias

from ovld import Medley, call_next, ovld, recurse
from ovld.dependent import Regexp

from ..ctx import Context, Patch, Patcher
from ..instructions import Instruction, T
from ..priority import STD

try:
    from cryptography.fernet import Fernet
except ImportError:  # pragma: no cover
    raise ImportError(
        "The 'cryptography' package is required for encryption features. "
        "Please install it with 'pip install cryptography'."
    )


#############
# Constants #
#############


if TYPE_CHECKING:
    Secret: TypeAlias = Annotated[T, None]
else:
    Secret = Instruction("Secret")


crypt_prefix = "~CRYPT~"


###########
# Context #
###########


class EncryptionKey(Context):
    password: str | Callable = None
    fernet_key: Fernet = None

    def __post_init__(self):
        if callable(self.password):
            self.password = cache(self.password)

    def get_encryption_key(self):
        if self.fernet_key is None:
            pw = self.password
            if callable(pw):
                pw = self.password()
            elif pw is None:
                raise Exception("No encryption password or key was provided")
            encoded = base64.b64encode(sha256(pw.encode()).digest())
            self.fernet_key = Fernet(encoded)

        return self.fernet_key

    def encrypt(self, value):
        ek = self.get_encryption_key()
        encrypted = ek.encrypt(json.dumps(value).encode("utf8")).decode("utf8")
        return f"{crypt_prefix}{encrypted}"

    def decrypt(self, encrypted: str):
        ek = self.get_encryption_key()
        return json.loads(ek.decrypt(encrypted.lstrip(crypt_prefix)))


##################
# Implementation #
##################

PRIO1 = STD.next()
PRIO2 = STD.next()


class Encrypt(Medley):
    @ovld(priority=PRIO2)
    def serialize(self, t: type[Any @ Secret], obj: object, ctx: EncryptionKey):
        result = call_next(Secret.strip(t), obj, ctx - EncryptionKey)
        return ctx.encrypt(result)

    @ovld(priority=PRIO2)
    def deserialize(
        self, t: type[Any @ Secret], obj: Regexp[rf"^{crypt_prefix}.*"], ctx: EncryptionKey
    ):
        obj = ctx.decrypt(obj)
        return call_next(Secret.strip(t), obj, ctx - EncryptionKey)

    @ovld(priority=PRIO1)
    def deserialize(self, t: type[Any @ Secret], obj: Any, ctx: EncryptionKey + Patcher):
        def crypt():
            return ctx.encrypt(obj)

        ctx.declare_patch(Patch(crypt, description="Encrypt value"))
        return recurse(Secret.strip(t), obj, ctx)



import json
from pathlib import Path

from ovld import ovld, recurse

from ..ctx import Location
from ..utils import import_any
from .abc import FileFormat

yaml = import_any(
    feature="YAML loading and dumping",
    candidates={
        "pyyaml:yaml": lambda m: m,
    },
)

Loader = getattr(yaml, "CSafeLoader", yaml.SafeLoader)
Dumper = getattr(yaml, "CSafeDumper", yaml.SafeDumper)


def yaml_source_extract(node, origin):
    return Location(
        source=origin,
        start=node.start_mark.index,
        end=node.end_mark.index,
        linecols=(
            (node.start_mark.line, node.start_mark.column),
            (node.end_mark.line, node.end_mark.column),
        ),
    )


@ovld
def locate(obj: yaml.MappingNode, origin: Path, trail: tuple | list):
    if trail:
        nxt, *rest = trail
        for k, v in obj.value:
            if k.value == nxt:
                return recurse(v, origin, rest)
    return yaml_source_extract(obj, origin)


@ovld
def locate(obj: yaml.SequenceNode, origin: Path, trail: tuple | list):
    if trail:
        nxt, *rest = trail
        for i, v in enumerate(obj.value):
            if i == nxt:
                return recurse(v, origin, rest)
    return yaml_source_extract(obj, origin)  # pragma: no cover


@ovld
def locate(obj: yaml.ScalarNode, origin: Path, trail: tuple | list):
    return yaml_source_extract(obj, origin)


class YAML(FileFormat):
    def locate(self, f: Path, trail: tuple[str]):
        return locate(yaml.compose(f.read_text(), Loader), f, trail)

    def patch(self, source, patches):
        for start, end, content in sorted(patches, reverse=True):
            source = source[:start] + json.dumps(content) + source[end:]
        return source

    def loads(self, s: str):
        return yaml.load(s or "{}", Loader)

    def dumps(self, data):
        return yaml.dump(data, Dumper=Dumper, allow_unicode=True, sort_keys=False)



import pytest

from serieux.formats import dump, dumps, load, loads

data = {
    "plums": 38,
    "hello": "world",
    "things": [1, 2, 3.5, True],
}

cases = [
    ("json", data),
    ("pkl", data),
    ("yaml", data),
    ("toml", data),
    ("txt", "hello!"),
]


@pytest.mark.parametrize("suffix, value", cases)
def test_dump_and_load(tmp_path, suffix, value):
    file = tmp_path / f"test.{suffix}"
    dump(file, value, suffix)
    loaded = load(file, suffix)
    assert loaded == value


str_cases = [
    ("json", data),
    ("yaml", data),
    ("toml", data),
    ("txt", "hello!"),
]


@pytest.mark.parametrize("suffix, value", str_cases)
def test_dumps_and_loads(suffix, value):
    dumped = dumps(value, suffix)
    loaded = loads(dumped, suffix)
    assert loaded == value



import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Literal

import pytest

from serieux import deserialize
from serieux.ctx import Trail, empty
from serieux.exc import ValidationError, display
from serieux.model import AllowExtras

from .common import has_312_features, one_test_per_assert
from .definitions import (
    Character,
    Color,
    Defaults,
    DIDHolder,
    DotDict,
    IdentifiedCar,
    Level,
    LTHolder,
    Point,
    Point3D,
    Pointato,
)

here = Path(__file__).parent


@one_test_per_assert
def test_deserialize_scalars():
    assert deserialize(int, 0) == 0
    assert deserialize(int, 12) == 12
    assert deserialize(float, -3.25) == -3.25
    assert deserialize(str, "flagada") == "flagada"
    assert deserialize(bool, True) is True
    assert deserialize(bool, False) is False
    assert deserialize(type(None), None) is None


@one_test_per_assert
def test_deserialize_scalars_conversion():
    assert deserialize(float, 10) == 10.0


@one_test_per_assert
def test_deserialize_object():
    assert deserialize(object, 10) == 10
    assert deserialize(Any, 10) == 10


def test_deserialize_dict():
    assert deserialize(dict[str, int], {"a": 1, "b": 2}) == {"a": 1, "b": 2}


def test_deserialize_point():
    assert deserialize(Point, {"x": 1, "y": 2}) == Point(1, 2)


def test_deserialize_list_of_points():
    pts = [
        {"x": 1, "y": 2},
        {"x": 3, "y": 4},
    ]
    assert deserialize(list[Point], pts) == [Point(1, 2), Point(3, 4)]


def test_deserialize_dict_of_points():
    pts = {
        "pt1": {"x": 1, "y": 2},
        "pt2": {"x": 3, "y": 4},
    }
    assert deserialize(dict[str, Point], pts) == {
        "pt1": Point(1, 2),
        "pt2": Point(3, 4),
    }


def test_deserialize_set():
    nums = [1, 4, 2, 3, 3, 4]
    assert deserialize(set[int], nums) == {1, 2, 3, 4}


def test_deserialize_frozenset():
    nums = [1, 4, 2, 3, 3, 4]
    assert deserialize(frozenset[int], nums) == frozenset({1, 2, 3, 4})


def test_deserialize_dict_subclass():
    data = {"apple": 3, "banana": 7}
    deser = deserialize(DotDict[str, int], data)
    assert deser.apple == 3
    assert deser.banana == 7


@one_test_per_assert
def test_deserialize_union():
    assert deserialize(str | int, 3) == 3
    assert deserialize(str | int, "wow") == "wow"
    assert deserialize(Point | int, 3) == 3


@dataclass
class Poink:
    x: int
    y: int


def test_cannot_deserialize_undistinguishable():
    with pytest.raises(Exception, match="Cannot differentiate"):
        deserialize(Point | Poink, {"x": 1, "y": 2})


@has_312_features
def test_deserialize_tree():
    from .definitions_py312 import Tree

    tree = {
        "left": {
            "left": 1,
            "right": 2,
        },
        "right": {
            "left": {
                "left": {
                    "left": 3,
                    "right": 4,
                },
                "right": 5,
            },
            "right": 6,
        },
    }

    assert deserialize(Tree[int], tree) == Tree(Tree(1, 2), Tree(Tree(Tree(3, 4), 5), 6))


def test_deserialize_overlapping_union():
    P = Point | Point3D
    assert type(deserialize(P, {"x": 1, "y": 2})) is Point
    assert type(deserialize(P, {"x": 1, "y": 2, "z": 3})) is Point3D

    # Make sure it also works the other way around
    P = Point3D | Point
    assert type(deserialize(P, {"x": 1, "y": 2})) is Point
    assert type(deserialize(P, {"x": 1, "y": 2, "z": 3})) is Point3D


def test_deserialize_defaults():
    data1 = {"name": "bob"}
    data2 = {"cool": True, "name": "alice"}

    x1 = deserialize(Defaults, data1)
    assert not x1.cool

    x2 = deserialize(Defaults, data2)
    assert x2.cool

    assert not x1.aliases
    assert not x2.aliases
    assert x1.aliases is not x2.aliases


def test_deserialize_enum():
    assert deserialize(Color, "red") == Color.RED
    assert deserialize(list[Color], ["green", "blue", "green"]) == [
        Color.GREEN,
        Color.BLUE,
        Color.GREEN,
    ]


def test_deserialize_enum_int():
    assert deserialize(Level, 1) == Level.MED
    assert deserialize(list[Level], [2, 0, 2]) == [Level.HI, Level.LO, Level.HI]


def test_deserialize_literal_enum():
    assert deserialize(Literal["red", "green", "blue"], "red") == "red"
    with pytest.raises(
        ValidationError,
        match=r"'yellow' is not a valid option for typing.Literal\['red', 'green', 'blue'\]",
    ):
        deserialize(Literal["red", "green", "blue"], "yellow")


def test_deserialize_literal_int_enum():
    assert deserialize(Literal[1, 2, 7], 2) == 2
    with pytest.raises(
        ValidationError,
        match=r"'3' is not a valid option for typing.Literal\[1, 2, 7\]",
    ):
        deserialize(Literal[1, 2, 7], 3)


def test_deserialize_literal_mixed_enum():
    lit = Literal[1, True, "quack"]
    assert deserialize(lit, 1) == 1
    assert deserialize(lit, True) is True
    assert deserialize(lit, "quack") == "quack"
    with pytest.raises(ValidationError):
        deserialize(lit, 3)
    with pytest.raises(ValidationError):
        deserialize(lit, "boop")
    with pytest.raises(ValidationError):
        deserialize(lit, False)


###############
# Error tests #
###############


def test_deserialize_scalar_error():
    with pytest.raises(ValidationError, match=r"Cannot deserialize string 'foo'"):
        deserialize(int, "foo")


def test_deserialize_scalar_error_2():
    with pytest.raises(ValidationError, match=r"Cannot deserialize object `13`"):
        deserialize(str, 13)


@pytest.mark.parametrize("ctx", (Trail(), empty))
def test_deserialize_missing_field(ctx):
    pts = [
        {"x": 1, "y": 2},
        {"x": 3},
    ]
    with pytest.raises(ValidationError, match=r"At path .1: Missing required field 'y'"):
        deserialize(list[Point], pts, ctx)


def test_deserialize_extra_fields_not_allowed():
    data = {"x": 1, "y": 2, "poop": 123}
    with pytest.raises(ValidationError, match=r"Extra unrecognized fields.*Point.*poop"):
        deserialize(Point, data)


def test_deserialize_extra_fields_allowed():
    data = {"x": 1, "y": 2, "poop": 123}
    assert deserialize(AllowExtras[Point], data) == Point(1, 2)


def test_deserialize_extra_fields_allowed_in_config():
    data = {
        "name": "Billy",
        "age": 28,
        "occupation": "Wizard",
        "backstory": "Mysterious",
        "splat": 99,
        "splatity_splat": True,
    }
    assert deserialize(Character, data) == Character(
        name="Billy", age=28, occupation="Wizard", backstory="Mysterious"
    )


def test_error_display(capsys, file_regression):
    pts = [
        {"x": 1, "y": 2},
        {"x": 3},
    ]
    with pytest.raises(ValidationError, match=r"At path .1: Missing required field 'y'") as exc:
        deserialize(list[Point], pts)

    display(exc.value, sys.stderr)
    cap = capsys.readouterr()
    file_regression.check("\n".join([cap.out, "=" * 80, cap.err]))


def test_deserialize_recursive_type():
    data = {
        "did": {
            "a": 1,
            "b": {
                "c": 2,
                "d": 3,
            },
        }
    }
    deser = deserialize(DIDHolder, data)
    assert deser.did == data["did"]


def test_deserialize_recursive_type_2():
    def p(x):
        return Point(x, x)

    data = {"lt": [{"x": 1, "y": 1}, [{"x": 2, "y": 2}, [{"x": 3, "y": 3}], {"x": 4, "y": 4}]]}
    deser = deserialize(LTHolder, data)
    assert deser.lt == [p(1), [p(2), [p(3)], p(4)]]


@pytest.mark.skipif(sys.version_info < (3, 12), reason="Requires Python 3.12+")
def test_deserialize_recursive_type_py312():
    from .definitions_py312 import LTHolder, Point

    def p(x):
        return Point(x, x)

    data = {"lt": [{"x": 1, "y": 1}, [{"x": 2, "y": 2}, [{"x": 3, "y": 3}], {"x": 4, "y": 4}]]}
    deser = deserialize(LTHolder, data)
    assert deser.lt == [p(1), [p(2), [p(3)], p(4)]]


class Blooper:
    def __init__(self, x: int, y: int, txt: str):
        self.message = txt * (x + y)


def test_deserialize_from_init():
    data = {"x": 2, "y": 3, "txt": "ha"}
    blooper = deserialize(Blooper, data)
    assert isinstance(blooper, Blooper)
    assert blooper.message == "hahahahaha"


def test_deserialize_pointato():
    with pytest.raises(ValidationError, match="Did you mean for it to be a dataclass"):
        deserialize(Pointato, {"x": 1, "y": 2})


def test_deserialize_kwonly_fields():
    data = {"id": 42, "horsepower": 300}
    car = deserialize(IdentifiedCar, data)
    assert car == IdentifiedCar(42, horsepower=300)



import getpass
import json
import os
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Mapping, Sequence

from . import CommandLineArguments, TaggedUnion, deserialize, schema, serieux
from .auto import Auto
from .ctx import Patcher, empty
from .features.encrypt import EncryptionKey
from .features.fromfile import IncludeFile
from .features.prompt import Promptable
from .features.registered import Referenced
from .model import field_at

srx = serieux + IncludeFile()


def value_at(data, path):
    for part in path.split("."):
        if isinstance(data, Sequence):
            data = data[int(part)]
        elif isinstance(data, Mapping):  # pragma: no cover
            data = data[part]
        else:
            data = getattr(data, part)
    return data


def model_at(model, path):
    model = field_at(model, path)
    if model is None:  # pragma: no cover
        sys.exit(f"No model found at {path!r}")
    return model.type


@dataclass
class Schema:
    """Dump the JSON schema of a class."""

    # A module:symbol reference for the schema
    # [positional]
    model: Referenced[Any]

    # Output file
    # [option: -o]
    out: Path = None

    def __call__(self):
        sch = schema(self.model).compile()
        txt = json.dumps(sch, indent=4)
        if self.out:
            self.out.write_text(txt)
        else:
            print(txt)


def enter_password():  # pragma: no cover
    return getpass.getpass("Enter password: ")


@dataclass(kw_only=True)
class FileOperation:
    # A module:symbol reference for the schema
    # [alias: -m]
    model: Referenced[Any]

    # Encryption password
    # [alias: -p]
    password: str = None

    # Input file
    # [alias: -f]
    file: Path

    def load(self, base_ctx=empty):
        ctx = Promptable() + base_ctx
        if self.password:
            ctx += EncryptionKey(password=self.password)
        elif "SERIEUX_PASSWORD" in os.environ:
            ctx += EncryptionKey(password=os.environ["SERIEUX_PASSWORD"])
        else:  # pragma: no cover
            ctx += EncryptionKey(password=enter_password)
        return srx.deserialize(self.model, self.file, ctx)


@dataclass(kw_only=True)
class SelectableFileOperation(FileOperation):
    # Path to select
    select: str = ""

    def load(self, base_ctx=empty):
        result = super().load(base_ctx)
        if self.select:
            result = value_at(result, self.select)
        return result

    def get_model(self):
        return model_at(self.model, self.select) if self.select else self.model


@dataclass(kw_only=True)
class Dump(SelectableFileOperation):
    """Dump configuration."""

    # Format to dump into
    format: str = "yaml"

    # File to dump into
    # [alias: -o]
    out: Path = None

    def __call__(self):
        m = self.get_model()
        result = self.load()
        serialized = srx.dump(m, result, dest=self.out, format=self.format)
        if serialized is not None:
            print(serialized.strip())


@dataclass(kw_only=True)
class Check(SelectableFileOperation):
    """Check configuration (true/false)."""

    def __call__(self):
        try:
            data = self.load()
        except (AttributeError, ValueError, KeyError):
            print("nonexistent")
            exit(2)

        if data:
            print("true")
            exit(0)
        elif not data:
            print("false")
            exit(1)


@dataclass(kw_only=True)
class Patch(FileOperation):
    """Patch a configuration file for prompts and secrets."""

    # Output file (will modify inplace if omitted)
    # [alias: -o]
    out: Path = None

    def __call__(self):
        patcher = Patcher()
        self.load(base_ctx=patcher)
        remap = {self.file: self.out} if self.out else None
        patcher.apply_patches(file_remap=remap)

        if patcher.patches:
            print("\033[1;36mThe following patches were applied:\033[0m")
            for p in patcher.patches.values():
                print(
                    f"\033[1;33m[{'.'.join(str(x) for x in p.ctx.trail)}]\033[0m \033[1;32m{p.description}\033[0m"
                )
        else:  # pragma: no cover
            print("\033[1;36mNo patches were applied.\033[0m")


@dataclass(kw_only=True)
class Run:
    """Run a function or class."""

    # Reference to the function or class to run
    # [positional]
    func: Referenced[Any]

    # Function arguments
    # [positional: ...]
    args: list[str]

    def __call__(self):
        result = deserialize(Auto[self.func], CommandLineArguments(self.args))
        if callable(result):
            result = result()
        if result is not None:
            print(result)


@dataclass
class SerieuxCommand:
    """Do things with serieux configurations."""

    # The command to run
    command: TaggedUnion[Schema, Dump, Check, Patch, Run]

    def __call__(self):  # pragma: no cover
        self.command()


def main(argv=None):  # pragma: no cover
    sys.path.insert(0, str(Path.cwd()))

    if argv is None:
        argv = sys.argv[1:]

    cmd = deserialize(SerieuxCommand, CommandLineArguments(arguments=argv))
    cmd()


if __name__ == "__main__":  # pragma: no cover
    main()



from functools import partial

from ..utils import import_any
from .abc import FileFormat

loads, dumps = import_any(
    feature="JSON loading and dumping",
    candidates={
        "msgspec.json": lambda m: (m.decode, m.encode),
        "orjson": lambda m: (m.loads, m.dumps),
        "ujson": lambda m: (m.loads, partial(m.dumps, ensure_ascii=False)),
        "json": lambda m: (m.loads, partial(m.dumps, ensure_ascii=False)),
    },
)


class JSON(FileFormat):
    def loads(self, s: str):
        return loads(s)

    def dumps(self, data):
        result = dumps(data)
        if isinstance(result, bytes):
            result = result.decode("utf-8")
        return result



from pathlib import Path


class FileFormat:  # pragma: no cover
    def locate(self, f: Path, trail: tuple[str]):
        return None

    def patch(self, source, patches):
        raise NotImplementedError(f"{type(self).__name__} does not implement `patch`")

    def loads(self, s: str):
        raise NotImplementedError(f"{type(self).__name__} does not implement `loads`")

    def dumps(self, data):
        raise NotImplementedError(f"{type(self).__name__} does not implement `dumps`")

    def load(self, f: Path):
        return self.loads(f.read_text())

    def dump(self, f: Path, data):
        f.write_text(self.dumps(data))

    @classmethod
    def serieux_from_string(cls, suffix):
        from . import registry

        return registry[suffix.lstrip(".")]



import os
from dataclasses import dataclass, field
from typing import get_args

from ovld import Medley, recurse
from ovld.dependent import Regexp

from serieux import Context, Serieux, deserialize
from serieux.exc import ValidationError

##################
# Implementation #
##################


@dataclass
class EnvValue:
    value: str


@dataclass
class EnvContext(Context):
    environ: dict[str, str] = field(default_factory=lambda: os.environ)


@Serieux.extend
class EnvInterpolator(Medley):
    """Custom serializer that interpolates environment variables in strings."""

    def deserialize(self, t: type[object], obj: Regexp[r"^\$[A-Z_][A-Z0-9_]*$"], ctx: EnvContext):
        var_name = obj.lstrip("$")
        value = ctx.environ.get(var_name)
        if value is None:
            raise ValidationError(f"Environment variable {var_name} not found")
        return recurse(t, EnvValue(value), ctx)

    def deserialize(self, t: type[str], obj: EnvValue, ctx: EnvContext):
        return obj.value

    def deserialize(self, t: type[bool], obj: EnvValue, ctx: EnvContext):
        value = obj.value.lower()
        if value in ("true", "1", "yes", "on"):
            return True
        elif value in ("false", "0", "no", "off"):
            return False
        raise ValidationError("Environment variable value cannot be converted to bool")

    def deserialize(self, t: type[int] | type[float], obj: EnvValue, ctx: EnvContext):
        try:
            return t(obj.value)
        except ValueError:
            raise ValidationError(
                f"Environment variable value cannot be converted to {t.__name__}"
            )

    def deserialize(self, t: type[list[object]], obj: EnvValue, ctx: EnvContext):
        element_type = get_args(t)[0]
        return [
            recurse(element_type, EnvValue(item.strip()), ctx) for item in obj.value.split(",")
        ]


#################
# Demonstration #
#################


def main():
    os.environ["DEBUG"] = "true"
    os.environ["PORT"] = "8080"
    os.environ["PI"] = "3.14159"
    os.environ["NAMES"] = "alice,bob,charlie"
    os.environ["FLOATS"] = "1.1,2.2,3.3"

    ctx = EnvContext()

    debug = deserialize(bool, "$DEBUG", ctx)
    print(f"Debug mode: {debug}")
    assert debug is True

    port = deserialize(int, "$PORT", ctx)
    print(f"Port: {port}")
    assert port == 8080

    pi = deserialize(float, "$PI", ctx)
    print(f"Pi: {pi}")
    assert abs(pi - 3.14159) < 0.00001

    names = deserialize(list[str], "$NAMES", ctx)
    print(f"Names: {names}")
    assert names == ["alice", "bob", "charlie"]

    numbers = deserialize(list[int], [1, 2, "$PORT"], ctx)
    print(f"Numbers: {numbers}")
    assert numbers == [1, 2, 8080]

    floats = deserialize(list[float], "$FLOATS", ctx)
    print(f"Floats: {floats}")
    assert floats == [1.1, 2.2, 3.3]


if __name__ == "__main__":
    main()



from dataclasses import dataclass, field, fields, make_dataclass, replace
from functools import reduce
from typing import TYPE_CHECKING, Annotated, Any, TypeAlias

from ovld import Medley, call_next, ovld, recurse

from ..ctx import Context
from ..exc import (
    NotGivenError,
    SerieuxError,
    ValidationError,
    ValidationExceptionGroup,
    merge_errors,
)
from ..instructions import Instruction, T, has_instruction, strip
from ..model import FieldModelizable, ListModelizable, Model, model
from ..priority import HI4
from .proxy import LazyProxy

#############
# Constants #
#############


if TYPE_CHECKING:
    Partial: TypeAlias = Annotated[T, None]
else:
    Partial = Instruction("Partial", annotation_priority=1, inherit=True)


class NOT_GIVEN_T:
    pass


NOT_GIVEN = NOT_GIVEN_T()


class PartialBase:
    pass


class AllTrails(Context):
    pass


class Sources:
    def __init__(self, *sources):
        self.sources = []
        for src in sources:
            if isinstance(src, Sources):  # pragma: no cover
                self.sources.extend(src.sources)
            else:
                self.sources.append(src)


@ovld
def partialize(t: type[FieldModelizable]):
    m = model(t)
    fields = [
        (
            f.name,
            Partial[f.type],
            field(default=NOT_GIVEN, metadata={"description": f.description, **f.metadata}),
        )
        for f in m.fields
    ]
    fields.append(
        (
            "_serieux_ctx",
            Context,
            field(repr=False, default=NOT_GIVEN, metadata={"serieux_metavar": "$ctx"}),
        )
    )
    dc = make_dataclass(
        cls_name=f"Partial[{t.__name__}]",
        bases=(PartialBase,),
        fields=fields,
        namespace={"_constructor": staticmethod(m.constructor), "_model": m},
    )
    return dc


@dataclass
class PartialListModelizable:
    elements: list


@ovld(priority=1)
def partialize(t: type[ListModelizable]):
    m = model(t)
    ef = m.element_field

    class PLM(PartialListModelizable):
        _model = m

        @classmethod
        def serieux_model(cls, call_next):
            return Model(
                original_type=cls,
                element_field=replace(ef, type=Partial[ef.type]),
                list_constructor=lambda xs: PLM(list(xs)),
            )

    return PLM


@ovld
def partialize(t: type[PartialBase]):  # pragma: no cover
    return t


@ovld
def partialize(t: object):
    return Partial[t]


###################
# Implementations #
###################


class PartialBuilding(Medley):
    @ovld(priority=HI4.next())
    def deserialize(self, t: type[Any @ Partial], obj: object, ctx: Context, /):
        try:
            return call_next(t, obj, ctx)
        except SerieuxError as exc:
            return exc
        except Exception as exc:
            return ValidationError(exc=exc)

    @ovld(priority=HI4.next())
    def deserialize(self, t: Any, obj: Sources, ctx: Context, /):
        is_partial = has_instruction(t, Partial)
        parts = []
        for src in obj.sources:
            try:
                parts.append(recurse(t if is_partial else Partial[t], src, ctx))
            except SerieuxError as exc:  # pragma: no cover
                parts.append(exc)
        merged = reduce(merge, parts)
        if is_partial:
            return merged
        rval = instantiate(merged)
        if isinstance(rval, SerieuxError):
            raise rval
        return rval

    @ovld(priority=HI4.next())
    def deserialize(self, t: Any, obj: object, ctx: AllTrails, /):
        data = obj if isinstance(obj, Sources) else Sources(obj)
        return recurse(t, data, ctx - AllTrails)


@model.register(priority=2)
def _(p: type[Any @ Partial]):
    return call_next(partialize(strip(p, Partial)))


######################
# Merge partial data #
######################


@ovld(priority=2)
def merge(x: object, y: NotGivenError):
    return x


@ovld(priority=2)
def merge(x: NotGivenError, y: object):  # pragma: no cover
    return y


@ovld(priority=2)
def merge(x: object, y: SerieuxError):
    return y


@ovld(priority=2)
def merge(x: SerieuxError, y: object):
    return x


@ovld(priority=3)
def merge(x: SerieuxError, y: SerieuxError):
    return ValidationExceptionGroup("Some errors occurred", [x, y])


@ovld(priority=1)
def merge(x: object, y: NOT_GIVEN_T):
    return x


@ovld(priority=1)
def merge(x: NOT_GIVEN_T, y: object):
    return y


@ovld(priority=1)
def merge(x: NOT_GIVEN_T, y: NOT_GIVEN_T):
    return NOT_GIVEN


@ovld
def merge(x: PartialBase, y: PartialBase):
    xm = x._model
    ym = y._model
    if xm is ym or xm.is_submodel_of(ym):
        main = type(x)
    elif ym.is_submodel_of(xm):
        main = type(y)
    else:
        raise ValidationError(
            f"Cannot merge sources because of incompatible constructors: '{xm}', '{ym}'"
        )
    args = {}
    for f in fields(main):
        xv = getattr(x, f.name, NOT_GIVEN)
        yv = getattr(y, f.name, NOT_GIVEN)
        args[f.name] = recurse(xv, yv)
    return main(**args)


@ovld
def merge(x: PartialBase, y: object):
    if (xc := x._model) is not model(type(y)):
        raise ValidationError(
            f"Cannot merge sources because of incompatible constructors: '{xc}', '{type(y)}'."
        )
    props = x._model.property_names
    kwargs = {k: v for k, v in vars(y).items() if k in props}
    return recurse(x, type(x)(**kwargs))


@ovld
def merge(x: object, y: PartialBase):
    if (yc := y._model) is not model(type(x)):
        raise ValidationError(
            f"Cannot merge sources because of incompatible constructors: '{type(x)}', '{yc}'."
        )
    props = y._model.property_names
    kwargs = {k: v for k, v in vars(x).items() if k in props}
    return recurse(type(y)(**kwargs), y)


@ovld
def merge(x: dict, y: dict):
    result = type(x)(x)
    for k, v in y.items():
        result[k] = recurse(result.get(k, NOT_GIVEN), v)
    return result


@ovld
def merge(x: PartialListModelizable, y: PartialListModelizable):
    assert x._model is y._model
    return type(x)(x.elements + y.elements)


@ovld
def merge(x: list, y: list):
    return x + y


@ovld
def merge(x: LazyProxy, y: LazyProxy):
    return LazyProxy(lambda: recurse(x._obj, y._obj), x._type)


@ovld
def merge(x: object, y: object):
    return y


############################
# Instantiate partial data #
############################


@ovld
def instantiate(xs: list):
    rval = []
    err = None
    for v in xs:
        value = recurse(v)
        if isinstance(value, SerieuxError):
            err = merge_errors(err, value)
        else:
            rval.append(value)
    return err if err else rval


@ovld
def instantiate(xs: dict):
    rval = type(xs)()
    err = None
    for k, v in xs.items():
        if v is NOT_GIVEN:
            continue
        value = recurse(v)
        if isinstance(value, SerieuxError):
            err = merge_errors(err, value)
        else:
            rval[k] = value
    return err if err else rval


@ovld
def instantiate(p: PartialBase):
    dc = p._constructor
    args = recurse({f.name: getattr(p, f.name) for f in p._model.fields})
    if isinstance(args, SerieuxError):
        return args
    try:
        return dc(**args)
    except Exception as exc:
        return ValidationError(exc=exc, ctx=p._serieux_ctx)


@ovld
def instantiate(p: PartialListModelizable):
    elems = recurse(p.elements)
    if isinstance(elems, SerieuxError):
        return elems
    return p._model.list_constructor(elems)


@ovld
def instantiate(x: LazyProxy):
    def do():
        rval = recurse(x._obj)
        if isinstance(rval, SerieuxError):
            raise rval
        return rval

    return LazyProxy(do, x._type)


@ovld
def instantiate(x: object):
    return x



import json
import shutil
from dataclasses import dataclass

import pytest

from serieux import Serieux
from serieux.ctx import Patcher
from serieux.features.encrypt import Encrypt, EncryptionKey, Secret, crypt_prefix

featured = (Serieux + Encrypt)()
serialize = featured.serialize
deserialize = featured.deserialize


@pytest.fixture
def ekey():
    return EncryptionKey(password="sesame")


@pytest.fixture
def bad_ekey():
    return EncryptionKey(password="peanuts")


@pytest.fixture
def no_ekey():
    return EncryptionKey(password=None)


@dataclass
class User:
    name: str
    password: Secret[str]


@dataclass
class Account:
    username: str
    secrets: Secret[dict[str, str]]
    public_info: str


def test_basic_encryption_decryption(ekey):
    original_value = "secret_data"
    encrypted = ekey.encrypt(original_value)

    assert encrypted.startswith(crypt_prefix)
    assert original_value not in encrypted

    decrypted = ekey.decrypt(encrypted)
    assert decrypted == original_value


def test_secret_string_serialization(ekey):
    user = User(name="alice", password="super_secret")
    serialized = serialize(User, user, ekey)

    assert serialized["name"] == "alice"
    assert serialized["password"].startswith(crypt_prefix)
    assert "super_secret" not in json.dumps(serialized)

    deserialized = deserialize(User, serialized, ekey)
    assert deserialized == user


def test_secret_dict_serialization(ekey):
    account = Account(
        username="bob",
        secrets={"api_key": "abc123", "token": "xyz789"},
        public_info="This is public",
    )

    serialized = serialize(Account, account, ekey)

    assert serialized["username"] == "bob"
    assert serialized["public_info"] == "This is public"
    assert serialized["secrets"].startswith(crypt_prefix)

    serialized_str = json.dumps(serialized)
    assert "abc123" not in serialized_str
    assert "xyz789" not in serialized_str

    deserialized = deserialize(Account, serialized, ekey)
    assert deserialized == account


def test_wrong_password_decryption(ekey, bad_ekey):
    user = User(name="charlie", password="my_secret")

    serialized = serialize(User, user, ekey)

    # Try to deserialize with wrong password - should fail
    with pytest.raises(Exception):
        deserialize(User, serialized, bad_ekey)


def test_no_password_decryption(ekey, no_ekey):
    user = User(name="charlie", password="my_secret")

    serialized = serialize(User, user, ekey)

    with pytest.raises(Exception):
        deserialize(User, serialized, no_ekey)


def test_callable_password():
    calls = [0]

    def get_pw():
        calls[0] += 1
        return "letmein"

    users = [
        User(name="Åke", password="hemlighet1"),
        User(name="Linnéa", password="blomma2"),
        User(name="Gösta", password="skog3"),
        User(name="Måns", password="sol4"),
        User(name="Sigrid", password="norrsken5"),
    ]

    serialized = serialize(list[User], users, EncryptionKey(password=get_pw))
    assert not any(u.password in str(serialized) for u in users)
    assert calls == [1]


def test_missing_encryption_context():
    user = User(name="dave", password="secret")
    serialized = serialize(User, user)
    assert serialized["name"] == "dave"
    assert serialized["password"] == "secret"


def test_invalid_encrypted_data(ekey):
    """Test deserialization with invalid encrypted data."""
    invalid_data = {"name": "eve", "password": f"{crypt_prefix}invalid_encrypted_data"}
    with pytest.raises(Exception):
        deserialize(User, invalid_data, ekey)


def test_encryption_with_patcher(ekey, datapath, tmp_path):
    src = datapath / "userpass.yaml"
    dest = tmp_path / "userpass.yaml"
    shutil.copy(src, dest)

    patcher = Patcher()
    ctx = ekey + patcher
    user = deserialize(User, dest, ctx)
    assert user.name == "frank"
    assert user.password == "plain_text_secret"
    assert len(patcher.patches) > 0

    assert crypt_prefix not in dest.read_text()
    patcher.apply_patches()
    assert crypt_prefix in dest.read_text()

    patcher = Patcher()
    ctx = ekey + patcher
    user = deserialize(User, dest, ctx)
    assert user.name == "frank"
    assert user.password == "plain_text_secret"
    assert len(patcher.patches) == 0



from ovld import Exactly, Medley, ovld, recurse

from ..ctx import Context
from ..priority import LO1
from .tagset import ReferencedClass

###################
# Implementations #
###################


class AutoTagAny(Medley):
    @ovld(priority=LO1)
    def serialize(self, t: type[Exactly[object]], obj: object, ctx: Context, /):
        return recurse(t @ ReferencedClass, obj, ctx)

    @ovld(priority=LO1)
    def deserialize(self, t: type[Exactly[object]], obj: dict, ctx: Context, /):
        return recurse(t @ ReferencedClass, obj, ctx)



import json
import os
import re
from dataclasses import field
from pathlib import Path
from types import NoneType
from typing import Annotated, Any, Literal, get_args

from ovld import Medley, call_next, ovld, recurse

from ..ctx import Trail
from ..exc import NotGivenError, ValidationError
from ..instructions import strip
from ..priority import HI1
from ..utils import UnionAlias
from .lazy import LazyProxy
from .partial import Sources


@ovld
def decode_string(t: type[int] | type[float] | type[str], value: str):
    return t(value)


@ovld
def decode_string(t: type[NoneType], value: str):
    val = value.lower()
    if val in ("", "null", "none"):
        return None
    else:
        raise ValidationError(f"Cannot convert '{value}' to None")


@ovld
def decode_string(t: type[bool], value: str):
    val = value.lower()
    if val in ("true", "1", "yes", "on"):
        return True
    elif val in ("false", "0", "no", "off", ""):
        return False
    else:
        raise ValidationError(f"Cannot convert '{value}' to boolean")


@ovld
def decode_string(t: type[object], value: str):
    try:
        return json.loads(value)
    except json.JSONDecodeError:
        return value


@ovld
def decode_string(t: type[UnionAlias], value: str):
    err = None
    for opt in get_args(t):
        try:
            return decode_string(opt, value)
        except Exception as exc:
            err = exc
    raise err


@ovld
def decode_string(t: type[list], value: str):
    (element_type,) = get_args(t) or (object,)
    return [recurse(element_type, item.strip()) for item in str(value).split(",")]


@ovld
def decode_string(t: type[Annotated], value: str):
    return recurse(strip(t), value)


class Environment(Trail):
    refs: dict[tuple[str, ...], object] = field(default_factory=dict, repr=False)
    environ: dict = field(default_factory=lambda: os.environ, repr=False)
    interpolation_pattern: re.Pattern = re.compile(r"\$\{([^}]+)\}")

    def __post_init__(self):
        if isinstance(self.interpolation_pattern, str):
            self.interpolation_pattern = re.compile(self.interpolation_pattern)

    def evaluate_reference(self, ref):
        def try_int(x):
            try:
                return int(x)
            except ValueError:
                return x

        stripped = ref.lstrip(".")
        dots = len(ref) - len(stripped)
        root = () if not dots else self.trail[:-dots]
        parts = [try_int(x) for x in stripped.split(".")]
        return self.refs[(*root, *parts)]

    @ovld
    def resolve_variable(self, t: Any, expr: str, /):
        match expr.split(":", 1):
            case (method, expr):
                return recurse(t, method, expr)
            case _:
                return recurse(t, "", expr)

    def resolve_variable(self, t: Any, method: Literal[""], expr: str, /):
        return LazyProxy(lambda: self.evaluate_reference(expr))

    def resolve_variable(self, t: Any, method: Literal["env"], expr: str, /):
        try:
            env_value = self.environ[expr]
        except KeyError:
            raise NotGivenError(f"Environment variable '{expr}' is not defined")
        else:
            return decode_string(t, env_value)

    def resolve_variable(self, t: Any, method: Literal["envfile"], expr: str, /):
        try:
            pth = Path(self.environ[expr]).expanduser()
        except KeyError:
            raise NotGivenError(f"Environment variable '{expr}' is not defined")
        if pth.exists():
            return pth
        else:
            return Sources(*[Path(x.strip()).expanduser() for x in str(pth).split(",")])

    def resolve_variable(self, t: Any, method: str, expr: str, /):
        raise ValidationError(
            f"Cannot resolve '{method}:{expr}' because the '{method}' resolver is not defined."
        )

    def __setitem__(self, pth, value):
        if not isinstance(pth, tuple):
            pth = (pth,)
        self.refs[pth] = value


class Interpolation(Medley):
    @ovld(priority=HI1(3))
    def deserialize(self, t: Any, obj: object, ctx: Environment):
        rval = call_next(t, obj, ctx)
        ctx.refs[ctx.trail] = rval
        return rval

    @ovld(priority=HI1(2))
    def deserialize(self, t: Any, obj: str, ctx: Environment):
        match ctx.interpolation_pattern.split(obj):
            case [s]:
                return call_next(t, s, ctx)
            case ["", expr, ""]:
                obj = ctx.resolve_variable(t, expr)
                if isinstance(obj, LazyProxy):

                    def interpolate():
                        return recurse(t, obj._obj, ctx)

                    return LazyProxy(interpolate)
                else:
                    return recurse(t, obj, ctx)
            case parts:

                def interpolate():
                    resolved = [
                        p if i % 2 == 0 else ctx.resolve_variable(str, p)
                        for i, p in enumerate(parts)
                    ]
                    subbed = "".join(map(str, resolved))
                    return recurse(t, subbed, ctx)

                return LazyProxy(interpolate)



from dataclasses import fields
from numbers import Number
from typing import TypeVar, Union

from serieux.utils import JSON, JSONLike, evaluate_hint as eh

from .common import has_312_features, one_test_per_assert
from .definitions import Point

T1 = TypeVar("T1")
T2 = TypeVar("T2")


@one_test_per_assert
def test_evaluate_hint():
    assert eh("str") is str
    assert eh(list["Point"], Point) == list[Point]
    assert eh(Union[int, "str"]) == int | str
    assert eh("int | str") == int | str


@one_test_per_assert
def test_evaluate_hint_generics():
    assert eh(dict[T1, T2]) == dict[T1, T2]
    assert eh(dict[T1, T2], typesub={T1: int}) == dict[int, T2]
    assert eh(dict[T1, T2], typesub={T2: int}) == dict[T1, int]
    assert eh(dict[T1, T2], typesub={T1: int, T2: str}) == dict[int, str]
    assert eh(dict[T2, T1], typesub={T1: int, T2: str}) == dict[str, int]


def test_evaluate_hint_tree():
    from .definitions import Tree

    for field in fields(Tree):
        assert eh(field.type, Tree) == Number | Tree


@has_312_features
def test_evaluate_hint_tree_parametric():
    from .definitions_py312 import Tree

    for field in fields(Tree):
        assert eh(field.type, Tree[float]) == Union[float, Tree[float]]
        assert eh(field.type, Tree[str]) == Union[str, Tree[str]]


def test_json_like():
    J = JSONLike[object]
    JL = JSONLike[list]
    for yes in [int, float, str, list[float], dict[str, str], dict[str, str | int]]:
        assert issubclass(yes, J)
    for no in [object, Point, dict[int, str], list]:
        assert not issubclass(no, J)
    assert not issubclass(dict[str, str], JL)


@one_test_per_assert
def test_json():
    assert isinstance(1, JSON)
    assert isinstance(3.14, JSON)
    assert isinstance("hello", JSON)
    assert isinstance([1, 2, 3], JSON)
    assert isinstance({"a": 1, "b": [2, {"z": 3}]}, JSON)
    assert isinstance([True, False, None], JSON)


class NotJson:
    pass


@one_test_per_assert
def test_not_json():
    assert not isinstance({"a": NotJson()}, JSON)
    assert not isinstance([1, 2, NotJson(), 3], JSON)
    assert not isinstance(NotJson(), JSON)
    assert not isinstance(set([1, 2, 3]), JSON)
    assert not isinstance((1, 2, 3), JSON)
    assert not isinstance(object(), JSON)



from .autotag import AutoTagAny
from .clargs import FromArguments
from .comment import CommentedObjects
from .dotted import DottedNotation
from .fromfile import FromFile, IncludeFile
from .interpol import Interpolation
from .lazy import LazyDeserialization
from .partial import PartialBuilding
from .tagset import TagSetFeature

__all__ = [
    "AutoTagAny",
    "CommentedObjects",
    "DottedNotation",
    "FromArguments",
    "FromFile",
    "IncludeFile",
    "Interpolation",
    "LazyDeserialization",
    "PartialBuilding",
    "TagSetFeature",
]



import os
from dataclasses import dataclass
from datetime import date
from pathlib import Path
from types import NoneType
from unittest import mock

import pytest

from serieux import Serieux
from serieux.exc import NotGivenError, ValidationError
from serieux.features.interpol import Environment, Interpolation
from serieux.features.partial import Sources
from serieux.instructions import Instruction
from tests.definitions import Country

deserialize = (Serieux + Interpolation)().deserialize

datapath = Path(__file__).parent.parent / "data"


@dataclass
class Player:
    name: str
    nickname: str
    number: int


@dataclass
class Team:
    name: str
    rank: int
    forward: Player
    defender: Player
    goalie: Player


def test_simple_interpolate():
    data = {"name": "Robert", "nickname": "${name}", "number": 1}
    assert deserialize(Player, data, Environment()) == Player(
        name="Robert",
        nickname="Robert",
        number=1,
    )


def test_relative():
    data = {
        "name": "Team ${forward.nickname}",
        "rank": 7,
        "forward": {"name": "Igor", "nickname": "${.name}", "number": 1},
        "defender": {"name": "Robert", "nickname": "${.name}${.name}", "number": 2},
        "goalie": {"name": "Harold", "nickname": "Roldy", "number": "${..rank}"},
    }
    assert deserialize(Team, data, Environment()) == Team(
        name="Team Igor",
        rank=7,
        forward=Player(name="Igor", nickname="Igor", number=1),
        defender=Player(name="Robert", nickname="RobertRobert", number=2),
        goalie=Player(name="Harold", nickname="Roldy", number=7),
    )


def test_chain():
    data = [
        {"name": "Aaron", "nickname": "Ho", "number": 1},
        {"name": "Barbara", "nickname": "${0.nickname}s", "number": 2},
        {"name": "Cornelius", "nickname": "${1.nickname}s", "number": 3},
        {"name": "Dominic", "nickname": "${2.nickname}s", "number": 4},
    ]
    players = deserialize(list[Player], data, Environment())
    assert str(players[1].nickname) == "Hos"
    assert str(players[2].nickname) == "Hoss"
    assert str(players[3].nickname) == "Hosss"


def test_refer_to_object():
    data = [{"name": "Jon", "nickname": "Pork", "number": 1}, "${0}"]
    players = deserialize(list[Player], data, Environment())
    assert players[0] == players[1]


@dataclass
class DateMix:
    sdate: str
    ddate: date


def test_further_conversion():
    data = {"sdate": "2025-05-01", "ddate": "${sdate}"}
    dm = deserialize(DateMix, data, Environment())
    assert dm.ddate == date(2025, 5, 1)


def test_further_conversion_2():
    data = {"sdate": "2025-05", "ddate": "${sdate}-01"}
    dm = deserialize(DateMix, data, Environment())
    assert dm.ddate == date(2025, 5, 1)


def test_deadlock():
    data = {"name": "${nickname}", "nickname": "${name}", "number": 1}
    player = deserialize(Player, data, Environment())
    with pytest.raises(Exception, match="Deadlock"):
        player.name == "x"


def test_env():
    with pytest.MonkeyPatch.context() as mp:
        mp.setenv("TESTOTRON", "123")
        data = {"name": "Jon", "nickname": "Jonathan", "number": "${env:TESTOTRON}"}
        player = deserialize(Player, data, Environment())
        assert player.name == "Jon"
        assert player.number == 123


def test_env_types():
    vars = Environment(environ={"BOOL": "yes"})
    vars = Environment(
        environ={
            "BOOL": "yes",
            "INT": "42",
            "FLOAT": "3.14",
            "STR": "hello",
            "LIST": "a,b,c",
            "BOOL_FALSE": "no",
            "DICT": '{"a": 1, "b": 2}',
            "EMPTY": "",
            "NULL": "null",
            "NONE": "NONE",
        }
    )

    assert deserialize(bool, "${env:BOOL}", vars) is True
    assert deserialize(int, "${env:INT}", vars) == 42
    assert deserialize(int | None, "${env:INT}", vars) == 42
    assert deserialize(int | None, "${env:EMPTY}", vars) is None
    assert deserialize(float, "${env:FLOAT}", vars) == 3.14
    assert deserialize(str, "${env:STR}", vars) == "hello"
    assert deserialize(list[str], "${env:LIST}", vars) == ["a", "b", "c"]
    assert deserialize(bool, "${env:BOOL_FALSE}", vars) is False
    assert deserialize(bool, "${env:EMPTY}", vars) is False
    assert deserialize(NoneType, "${env:EMPTY}", vars) is None
    assert deserialize(NoneType, "${env:NULL}", vars) is None
    assert deserialize(NoneType, "${env:NONE}", vars) is None
    assert deserialize(dict[str, int], "${env:DICT}", vars) == {"a": 1, "b": 2}


def test_env_custom_deser():
    from .test_usermeth import RGB

    vars = Environment(environ={"COLOR": "#ff00ff"})
    rgb = deserialize(RGB, "${env:COLOR}", vars)
    assert rgb == RGB(red=255, green=0, blue=255)


def test_invalid_boolean():
    with pytest.raises(ValidationError, match="Cannot convert 'invalid' to boolean"):
        deserialize(bool, "${env:INVALID}", Environment(environ={"INVALID": "invalid"}))


def test_invalid_null():
    with pytest.raises(ValidationError, match="Cannot convert 'invalid' to None"):
        deserialize(NoneType, "${env:INVALID}", Environment(environ={"INVALID": "invalid"}))


def test_invalid_union():
    with pytest.raises(ValidationError, match="Cannot convert 'invalid' to boolean"):
        deserialize(int | bool, "${env:INVALID}", Environment(environ={"INVALID": "invalid"}))


def test_unsupported_resolver():
    with pytest.raises(
        ValidationError,
        match="Cannot resolve 'unknown:xyz' because the 'unknown' resolver is not defined",
    ):
        deserialize(str, "${unknown:xyz}", Environment())


def test_not_given():
    with pytest.raises(NotGivenError, match="Environment variable 'MISSING' is not defined"):
        deserialize(str, "${env:MISSING}", Environment())


def test_annotated():
    X = Instruction("X")
    assert deserialize(X[str], "${env:XX}", Environment(environ={"XX": "nice"})) == "nice"


@dataclass
class Fool:
    name: str
    iq: int = 100


def test_not_given_ignore():
    srcs = Sources({"name": "John"}, {"iq": "${env:INTEL}"})

    d = deserialize(Fool, srcs, Environment())
    assert d.iq == 100

    d = deserialize(Fool, srcs, Environment(environ={"INTEL": "31"}))
    assert d.iq == 31


_canada = str(datapath / "canada.yaml")
_france = str(datapath / "france.yaml")


@mock.patch.dict(os.environ, {"FILOU": _canada})
def test_resolve_envfile():
    canada = deserialize(Country, "${envfile:FILOU}", Environment())
    assert canada.capital == "Ottawa"


@mock.patch.dict(os.environ, {"FILOU": f"{_canada}, {_france}"})
def test_resolve_envfile_two_files():
    canada = deserialize(Country, "${envfile:FILOU}", Environment())
    assert canada.capital == "Ottawa"
    assert [c.name for c in canada.citizens] == ["Olivier", "Abraham", "Jeannot"]


def test_resolve_envfile_not_given():
    canada = deserialize(Country, Sources(Path(_canada), "${envfile:FILOU}"), Environment())
    assert canada.capital == "Ottawa"


def test_populate_environment():
    env = Environment()
    env["exclaim"] = "wow!"
    result = deserialize(str, "${exclaim} I like this", env)
    assert result == "wow! I like this"


def test_custom_interpolate_regex():
    data = {"name": "Robert", "nickname": "~name", "number": 1}
    assert deserialize(Player, data, Environment(interpolation_pattern=r"~([a-z]+)")) == Player(
        name="Robert",
        nickname="Robert",
        number=1,
    )



import hashlib
import logging
import uuid
from collections import defaultdict
from dataclasses import dataclass, field, replace
from functools import cached_property
from pathlib import Path
from typing import Any, Callable

from ovld.medley import ChainAll, KeepLast, Medley

from .formats.abc import FileFormat

logger = logging.getLogger(__name__)


class Context(Medley, default_combiner=KeepLast):
    follow = ChainAll()


class EmptyContext(Context):
    def __add__(self, other):
        return other

    def __radd__(self, other):  # pragma: no cover
        return other


class Trail(Context):
    full_trail: tuple = ()

    @property
    def trail(self):
        return tuple(k for _, _, k in self.full_trail)

    def follow(self, objt, obj, field):
        return replace(self, full_trail=(*self.full_trail, (objt, obj, field)))


@dataclass
class Location:
    source: Path
    start: int
    end: int
    linecols: tuple

    @cached_property
    def whole_text(self):
        return self.source.read_text()

    @property
    def text(self):  # pragma: no cover
        return self.whole_text[self.start : self.end]


class WorkingDirectory(Context):
    directory: Path = None

    def make_path_for(self, *, name=None, suffix=None, entropy=None):
        if name is None and entropy is not None:
            name = hashlib.md5(
                str(entropy).encode() if isinstance(entropy, str) else entropy
            ).hexdigest()
        if name is None:
            name = str(uuid.uuid4())
        pth = self.directory / name
        if suffix is not None:
            pth = pth.with_suffix(suffix)
        return pth

    def save_to_file(
        self, data: str | bytes = None, suffix=None, *, name=None, callback=None, entropy=None
    ):
        dest = self.make_path_for(entropy=entropy or data, suffix=suffix, name=name)
        dest.parent.mkdir(parents=True, exist_ok=True)
        if callback:
            callback(dest)
        else:
            if isinstance(data, str):
                mode = "w"
                encoding = "utf-8"
            else:
                mode = "wb"
                encoding = None
            with open(dest, mode=mode, encoding=encoding) as f:
                f.write(data)
        return str(dest.relative_to(self.directory))


class Sourced(WorkingDirectory):
    origin: Path = None
    format: FileFormat = None
    source_trail: tuple = ()

    def __post_init__(self):
        if self.directory is None:
            self.directory = self.origin.parent

    def compute_location(self, trail=None):
        if isinstance(self, Trail):
            trail = self.trail
            pfx = len(self.source_trail)
            if trail[:pfx] != self.source_trail:  # pragma: no cover
                return None
            trail = trail[pfx:]
            return self.format.locate(self.origin, trail)
        return None  # pragma: no cover


def locate(ctx):
    if isinstance(ctx, Sourced):
        return ctx.compute_location()
    else:
        return None


@dataclass
class Patch:
    data: Callable | Any
    ctx: Context = None
    description: str | None = None

    def __post_init__(self):
        if self.description is None:
            self.description = f"Set to: {self.data!r}"

    def compute(self):
        if callable(self.data):  # pragma: no cover
            return self.data()
        else:
            return self.data

    def __str__(self):  # pragma: no cover
        return f"Patch({self.description!r})"


class Patcher(Trail):
    patches: dict[int, tuple[Context, Any]] = field(default_factory=dict)

    def declare_patch(self, patch):
        if not isinstance(patch, Patch):
            patch = Patch(patch, ctx=self)
        elif not patch.ctx:  # pragma: no cover
            patch = replace(patch, ctx=self)
        if start := (loc := locate(patch.ctx)) and loc.start:
            self.patches[start] = patch

    def apply_patches(self, file_remap=None):
        codes = {}
        patches = defaultdict(list)
        for patch in self.patches.values():
            if loc := locate(patch.ctx):
                codes[loc.source] = loc.whole_text
                patches[loc.source].append((loc.start, loc.end, patch.compute()))
            else:  # pragma: no cover
                logger.warning(f"Cannot apply patch at a context without a location: `{patch}`")
        for file, blocks in patches.items():
            code = patch.ctx.format.patch(codes[file], blocks)
            if file_remap:
                file = file_remap[file]
            file.write_text(code)


empty = EmptyContext()



from dataclasses import dataclass

from serieux.docstrings import VariableDoc, get_attribute_docstrings, get_variable_data


def _attribute_docstrings(t):
    return {k: v.doc for k, v in get_attribute_docstrings(t).items()}


def _var_data(t):
    return {k: v.doc for k, v in get_variable_data(t).items()}


@dataclass
class CommentedFields:
    # The name
    # of the person
    name: str
    # The age in years
    age: int
    # Whether the person is active
    is_active: bool = False
    # Optional email address
    email: str | None = None


@dataclass
class DocsFields:
    name: str
    """The person's full name."""

    age: int
    """The person's age in years."""

    is_active: bool = False
    """Whether the person's account is currently active."""

    email: str | None = None
    """The person's email address, if provided."""


def test_commented_fields():
    results = _attribute_docstrings(CommentedFields)
    assert results == {
        "name": "The name\nof the person",
        "age": "The age in years",
        "is_active": "Whether the person is active",
        "email": "Optional email address",
    }


def test_docs_fields():
    results = _attribute_docstrings(DocsFields)
    assert results == {
        "name": "The person's full name.",
        "age": "The person's age in years.",
        "is_active": "Whether the person's account is currently active.",
        "email": "The person's email address, if provided.",
    }


def add(
    # x coordinate
    x: int,
    # y coordinate
    y: int,
):
    return x + y


def test_function_docs():
    results = _var_data(add)
    assert results == {
        "x": "x coordinate",
        "y": "y coordinate",
    }


def add_pos(
    # xx coordinate
    xx: int,
    /,
    # yy coordinate
    yy: int,
):
    return xx + yy


def test_function_docs_2():
    results = _var_data(add_pos)
    assert results == {
        "xx": "xx coordinate",
        "yy": "yy coordinate",
    }


class Thing:
    # The width
    # [positional]
    # [mystery: 3] [moo]
    width: int
    # The length
    # [metavar: len]
    length: int


def test_metadata():
    results = get_attribute_docstrings(Thing)
    assert results == {
        "width": VariableDoc(
            doc="The width", metadata={"positional": True, "mystery": 3, "moo": True}
        ),
        "length": VariableDoc(doc="The length", metadata={"metavar": "len"}),
    }



from serieux.tell import tells


def test_tells_int():
    assert tells(int, int) == set()
    assert tells(int, str) is None


def test_tells_float():
    assert tells(float, float) == set()
    assert tells(float, str) is None


def test_tells_bool():
    assert tells(bool, bool) == set()
    assert tells(bool, dict) is None



import pickle
from pathlib import Path

from .abc import FileFormat


class PKL(FileFormat):
    def load(self, f: Path):
        return pickle.loads(f.read_bytes())

    def dump(self, f: Path, data):
        f.write_bytes(pickle.dumps(data))



import importlib
from dataclasses import dataclass, field
from typing import Any

from ovld import Medley, ovld, recurse

from ..ctx import Context
from ..exc import ValidationError
from ..instructions import BaseInstruction
from ..model import Model, model
from ..priority import HI2


class BaseRegistry(BaseInstruction):  # pragma: no cover
    def from_symbol(self, item):
        raise NotImplementedError()

    def to_symbol(self, item):
        raise NotImplementedError()

    def iterate(self):
        raise NotImplementedError()

    def closed(self):
        return True

    def __class_getitem__(cls, item):
        return item @ cls()


@dataclass(eq=False)
class Registry(BaseRegistry):
    registry: dict[str, Any] = field(default_factory=dict)
    inverse_registry: dict[Any, str] = None
    inverse: bool = True

    def __post_init__(self):
        if self.inverse and self.inverse_registry is None and self.inverse:
            self.inverse_registry = {v: k for k, v in self.registry.items()}

    def register(self, item, value):
        self.registry[item] = value
        if self.inverse:
            self.inverse_registry[value] = item
        return value

    def from_symbol(self, item):
        if item in self.registry:
            return self.registry[item]
        else:
            raise ValidationError(
                f"{item!r} is not a registered option. Should be one of: {list(self.registry.keys())}"
            )

    def to_symbol(self, value):
        if self.inverse and value in self.inverse_registry:
            return self.inverse_registry[value]
        else:
            raise ValidationError(
                f"The value {value!r} is not registered under any name in the mapping"
            )

    def iterate(self):
        yield from self.registry.keys()


class Referenced(BaseRegistry):
    def from_symbol(self, item):
        mod_name, _, class_name = item.partition(":")
        mod = importlib.import_module(mod_name)
        return getattr(mod, class_name) if class_name else mod

    def to_symbol(self, value):
        if hasattr(value, "__qualname__"):
            return f"{value.__module__}:{value.__qualname__}"
        else:
            raise ValidationError(f"Cannot find a module:symbol reference for the value {value!r}")

    def iterate(self):  # pragma: no cover
        yield from []

    def closed(self):  # pragma: no cover
        return False


class AutoRegistered:
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if not hasattr(cls, "_registry"):
            cls._registry = Registry()

    def __init__(self, registered_name):
        self.registered_name = registered_name
        type(self)._registry.register(registered_name, self)


def auto_singleton(arg, /):
    def wrap(cls):
        if (
            not isinstance(cls, type)
            or not issubclass(cls, AutoRegistered)
            or AutoRegistered in cls.__bases__
        ):
            raise TypeError(
                "@singleton must wrap a class definition and the class must be"
                " a subclass of Registered, but not a direct subclass."
            )
        return cls(registered_name=registered_name)

    if isinstance(arg, str):
        registered_name = arg
        return wrap
    else:
        registered_name = arg.__name__.lower()
        return wrap(arg)


class RegisteredHandler(Medley):
    @ovld(priority=HI2)
    def schema(self, t: type[AutoRegistered], ctx: Context, /):
        return {"type": "string", "enum": list(t._registry.iterate())}


@model.register
def _(t: type[Any @ BaseRegistry]):  # noqa: F821
    rg = BaseRegistry.extract(t)
    if not rg:
        return None
    return Model(
        original_type=t,
        from_string=rg.from_symbol,
        to_string=rg.to_symbol,
    )


@model.register
def _(t: type[AutoRegistered]):  # noqa: F821
    if not hasattr(t, "_registry"):
        return None
    return recurse(t @ t._registry)



from dataclasses import dataclass
from datetime import date

import pytest
from ovld import Medley
from ovld.dependent import Regexp

from serieux import Serieux
from serieux.ctx import Context, Trail
from serieux.exc import SerieuxError, ValidationError
from serieux.features.partial import NOT_GIVEN, AllTrails, Partial, PartialBuilding, Sources

from ..common import validation_errors
from ..definitions import Defaults, Elf, Player, Point

load = (Serieux + PartialBuilding)().load


def test_partial():
    one = load(Partial[Point], {"x": 10})
    assert one.x == 10
    two = load(Partial[Point], {"y": 30})
    assert two.y == 30


def test_partial_error():
    value = load(Partial[Point], {"x": "oh", "y": "no"})
    assert isinstance(value.x, ValidationError)
    assert isinstance(value.y, ValidationError)


def test_two_sources():
    assert load(Point, Sources({"x": 1}, {"y": 2})) == Point(1, 2)


def test_three_sources():
    assert load(Point, Sources({"x": 1}, {"y": 2}, {"x": 3})) == Point(3, 2)


def test_complicated_partial():
    d = load(
        dict[str, Point | str],
        Sources(
            {"a": {"x": 1, "y": 2}},
            {"a": {"x": 3}},
            {"b": "wow"},
        ),
    )
    assert d == {"a": Point(3, 2), "b": "wow"}


def test_partial_on_partial_type():
    ppt = load(Partial[Point], Sources({"x": 1}))
    assert ppt.x == 1
    assert ppt.y is NOT_GIVEN


def test_merge_lists():
    li = load(
        list[int],
        Sources(
            [1, 2],
            [3, 4, 5],
        ),
    )
    assert li == [1, 2, 3, 4, 5]


def test_partial_incompatibility():
    with pytest.raises(ValidationError, match="incompatible constructors"):
        load(
            Point | Player,
            Sources({"x": 1, "y": 2}, {"first": "Joan", "last": "Ark", "batting": 0.7}),
        )


def test_partial_incompatibility_2():
    with pytest.raises(ValidationError, match="incompatible constructors"):
        load(Point | str, Sources({"x": 1, "y": 2}, "waa"))


def test_partial_incompatibility_3():
    with pytest.raises(ValidationError, match="incompatible constructors"):
        load(Point | str, Sources("waa", {"x": 1, "y": 2}))


@dataclass
class Climate:
    hot: bool
    sunny: bool


@dataclass
class City:
    name: str
    population: int
    climate: Climate


@dataclass
class Country:
    name: str
    capital: City


def test_nested():
    d = load(
        Country,
        Sources(
            {"name": "Canada"},
            {"capital": {"name": "Ottawa"}},
            {"capital": {"population": 800000}},
            {"capital": {"climate": {"hot": False}}},
            {"capital": {"climate": {"sunny": False}}},
        ),
    )
    assert d == Country(
        name="Canada",
        capital=City(
            name="Ottawa",
            population=800_000,
            climate=Climate(
                hot=False,
                sunny=False,
            ),
        ),
    )


@dataclass
class Positive:
    m: int
    n: int

    def __post_init__(self):
        if self.m <= 0:
            raise ValueError("NO")
        if self.n <= 0:
            raise ValueError("Get that worthless number outta here")


def test_error_at_construction():
    with pytest.raises(ValidationError, match="At path .1"):
        load(list[Positive], [{"m": 3, "n": 7}, Sources({"m": 1}, {"n": -3})], Trail())


def test_multiple_errors():
    msg = "Cannot deserialize string"
    with validation_errors({".0.y": msg, ".1.x": msg, ".1.y": msg}):
        load(list[Point], Sources([{"x": 23, "y": "crap"}, {"x": "oh", "y": "no"}]), Trail())


def test_multiple_errors_2():
    errs = {
        ".name": "Cannot deserialize object `123`",
        ".birthdate": "Invalid isoformat string",
        ".favorite_color": "Cannot deserialize object `False`",
    }
    with validation_errors(errs):
        load(Elf, Sources({"name": 123, "birthdate": "no", "favorite_color": False}))


def test_multiple_errors_display(check_error_display):
    with check_error_display():
        load(list[Point], Sources([{"x": 23, "y": "crap"}, {"x": "oh", "y": "no"}]), Trail())


def test_multiple_errors_display_all_trails(check_error_display):
    with check_error_display():
        load(list[Point], [{"x": 23, "y": "crap"}, {"x": "oh", "y": "no"}], AllTrails())


def test_partial_defaults():
    result = load(Defaults, Sources({"name": "Nicolas"}, {"cool": True}))
    assert result == Defaults(name="Nicolas", aliases=[], cool=True)


@dataclass
class RGB:
    red: int
    green: int
    blue: int


@Serieux.extend
class RGBSerializer(Medley):
    def deserialize(self, t: type[RGB], obj: Regexp[r"^#[0-9a-fA-F]{6}$"], ctx: Context):
        hex_str = obj.lstrip("#")
        red = int(hex_str[0:2], 16)
        green = int(hex_str[2:4], 16)
        blue = int(hex_str[4:6], 16)
        return RGB(red=red, green=green, blue=blue)


def test_merge_partial_with_object():
    assert load(RGB, Sources({"red": 0}, "#ffffff")) == RGB(255, 255, 255)
    assert load(RGB, Sources("#ffffff", {"red": 0})) == RGB(0, 255, 255)

    with pytest.raises(SerieuxError, match="Cannot deserialize string"):
        load(RGB, Sources("#fffffX", {"red": 0}))

    with pytest.raises(SerieuxError, match="Cannot deserialize"):
        load(RGB, Sources({"red": 0}, "#fffffX"))

    with pytest.raises(SerieuxError, match="Some errors occurred"):
        load(RGB, Sources("#fffffX", "#fffffX"))


def test_partial_string_modelizable():
    assert load(Partial[date], "2023-05-15") == date(2023, 5, 15)



import re
from dataclasses import MISSING, dataclass, field, fields, is_dataclass, replace
from datetime import date, datetime, timedelta
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Callable,
    Optional,
    TypeAlias,
    get_args,
    get_origin,
)
from zoneinfo import ZoneInfo

from ovld import Dataclass, Lambda, call_next, class_check, ovld, recurse, subclasscheck

from .docstrings import VariableDoc, get_attribute_docstrings
from .exc import ValidationError
from .instructions import Instruction, T, inherit, strip
from .utils import UnionAlias, clsstring, evaluate_hint

UNDEFINED = object()


if TYPE_CHECKING:
    AllowExtras: TypeAlias = Annotated[T, None]
else:
    AllowExtras = Instruction("AllowExtras", annotation_priority=1, inherit=True)


@class_check
def Modelizable(t):
    return isinstance(model(t), Model)


@class_check
def StringModelizable(t):
    return isinstance(m := model(t), Model) and m.from_string is not None


@class_check
def FieldModelizable(t):
    return isinstance(m := model(t), Model) and m.fields is not None


@class_check
def ListModelizable(t):
    return isinstance(m := model(t), Model) and m.element_field is not None


@dataclass(kw_only=True)
class Field:
    name: str
    type: type
    description: str = None
    metadata: dict[str, object] = field(default_factory=dict)
    default: object = UNDEFINED
    default_factory: Callable = UNDEFINED

    argument_name: str | int = UNDEFINED
    property_name: str = UNDEFINED
    serialized_name: str = UNDEFINED

    # Not implemented yet
    flatten: bool = False

    # Meta-variable to store in this field
    metavar: str = None

    def __post_init__(self):
        if self.property_name is UNDEFINED:
            self.property_name = self.name
        if self.argument_name is UNDEFINED:  # pragma: no cover
            self.argument_name = self.name
        if self.serialized_name is UNDEFINED:
            self.serialized_name = self.name
        if self.default is UNDEFINED:  # pragma: no cover
            self.default = MISSING
        if self.default_factory is UNDEFINED:
            self.default_factory = MISSING

    @property
    def required(self):
        return self.default is MISSING and self.default_factory is MISSING


@dataclass
class Model:
    original_type: type
    fields: list[Field] = None
    element_field: Field = None
    constructor: Callable = None
    list_constructor: Callable = None
    from_string: Callable = None
    to_string: Callable = None
    regexp: re.Pattern = None
    string_description: str = None
    allow_extras: bool = False

    def __post_init__(self):
        if isinstance(self.regexp, str):
            self.regexp = re.compile(self.regexp)
        if self.element_field is not None and self.list_constructor is None:  # pragma: no cover
            self.list_constructor = self.constructor

    def accepts(self, other):
        ot = strip(self.original_type)
        return subclasscheck(other, get_origin(ot) or ot)

    def is_submodel_of(self, other):
        # TODO: check that the fields are also the same
        return subclasscheck(self.original_type, other.original_type)

    def __str__(self):
        return f"Model({clsstring(self.original_type)})"

    @cached_property
    def property_names(self):
        return {f.property_name for f in self.fields}

    @cached_property
    def constructed_type(self):
        if isinstance(self.constructor, type):
            return self.constructor
        elif self.constructor is None:  # pragma: no cover
            raise TypeError(f"No constructor defined for {self}.")
        else:
            return_type = getattr(self.constructor, "__annotations__", {}).get("return", None)
            if return_type is None:
                raise TypeError(
                    f"Constructor for {self} does not have a return type annotation defined."
                )
            return return_type

    __repr__ = __str__


def constructed_type(t):
    match model(t):
        case None:
            return strip(t)
        case Model() as m:
            return m.constructed_type


_model_cache = {}
_premade = {}


def _take_premade(t):
    _model_cache[t] = _premade.pop(t)
    return _model_cache[t]


#########
# model #
#########


@ovld(priority=100)
def model(t: type[Any]):
    t = evaluate_hint(t)
    if t not in _model_cache:
        _premade[t] = Model(
            original_type=t,
            fields=[],
            constructor=None,
        )
        m = _model_cache[t] = call_next(t)
        if isinstance(cfg := getattr(t, "SerieuxConfig", None), type):
            if (ae := getattr(cfg, "allow_extras", None)) is not None:
                m.allow_extras = ae
    return _model_cache[t]


def safe_isinstance(obj, t):
    try:
        return isinstance(obj, t)
    except TypeError:  # pragma: no cover
        return False


@ovld
def model(dc: type[Dataclass]):
    def make_field(i, field):
        for target in reversed(dc.mro()):
            # Find where the field was defined
            if is_dataclass(target) and field in fields(target):
                break

        typ = evaluate_hint(field.type, target, None, tsub)
        if field.default is None and not safe_isinstance(field.default, typ):
            typ = Optional[typ]

        vardoc = attributes.get(field.name, None) or VariableDoc("", {})
        meta = {**field.metadata, **vardoc.metadata}
        if meta.get("ignore", False):
            return None

        return Field(
            name=field.name,
            description=meta.get("description", None) or vardoc.doc,
            type=typ,
            default=field.default,
            default_factory=field.default_factory,
            flatten=meta.get("flatten", False),
            metavar=meta.get("serieux_metavar", None),
            metadata=meta,
            argument_name=field.name if field.kw_only else i,
        )

    rval = _take_premade(dc)
    tsub = {}
    constructor = dc
    if (origin := get_origin(dc)) is not None:
        tsub = dict(zip(origin.__type_params__, get_args(dc)))
        constructor = origin

    attributes = get_attribute_docstrings(dc)

    _fields = [make_field(i, field) for i, field in enumerate(fields(constructor))]
    rval.fields = [f for f in _fields if f]
    rval.constructor = constructor
    return rval


@ovld
def model(sq: type[list] | type[set] | type[frozenset]):
    (et,) = get_args(sq) or [object]
    return Model(
        original_type=sq,
        element_field=Field(name="*", type=et),
        list_constructor=get_origin(sq) or sq,
    )


@ovld
def model(t: type[Any @ AllowExtras]):
    m = call_next(strip(t, AllowExtras))
    return m and replace(m, allow_extras=True)


@ovld(priority=-1)
def model(t: type[Annotated]):
    if t is Annotated:  # pragma: no cover
        # This is hit in Python <= 3.12
        return None
    m = recurse(strip(t))
    if m and m.fields is not None:
        return Model(
            original_type=m.original_type,
            fields=[replace(field, type=inherit(t, field.type)) for field in m.fields],
            constructor=m.constructor,
        )
    else:
        return m


@ovld
def model(t: type[ZoneInfo]):
    return Model(
        original_type=t,
        from_string=ZoneInfo,
        to_string=Lambda("$obj.key"),
    )


@ovld
def model(t: type[date] | type[datetime]):
    return Model(
        original_type=t,
        from_string=Lambda("$t.fromisoformat($obj)"),
        to_string=Lambda("$t.isoformat($obj)"),
    )


def _timedelta_to_string(obj: timedelta):
    """Serialize timedelta as Xs (seconds) or Xus (microseconds)."""
    seconds = int(obj.total_seconds())
    if obj.microseconds:
        return f"{seconds}{obj.microseconds:06}us"
    else:
        return f"{seconds}s"


def _timedelta_from_string(obj: str):
    """Deserialize a combination of days, hours, etc. as a timedelta."""
    units = {
        "d": "days",
        "h": "hours",
        "m": "minutes",
        "s": "seconds",
        "ms": "milliseconds",
        "us": "microseconds",
    }
    sign = 1
    if obj.startswith("-"):
        obj = obj[1:]
        sign = -1
    kw = {}
    parts = re.split(string=obj, pattern="([a-z ]+)")
    assert parts[-1] == ""
    for i in range(len(parts) // 2):
        n = parts[i * 2]
        unit = parts[i * 2 + 1].strip()
        assert unit in units
        try:
            kw[units[unit]] = float(n)
        except ValueError:
            raise ValidationError(f"Could not convert '{n}' ({units[unit]}) to float")
    return sign * timedelta(**kw)


@ovld
def model(t: type[timedelta]):
    return Model(
        original_type=t,
        from_string=_timedelta_from_string,
        to_string=_timedelta_to_string,
        regexp=r"^[+-]?([\d.]+[dhms]|[\d.]+ms|[\d.]+us)+$",
        string_description="A string such as 1d, 5h or 3d5h7s, ending in a unit. Valid units are d, h, m, s, ms, us.",
    )


@ovld
def model(t: type[re.Pattern]):
    return Model(
        original_type=t,
        from_string=re.compile,
        to_string=Lambda("$obj.pattern"),
    )


@ovld(priority=-1)
def model(t: object):
    return None


############
# field_at #
############


@ovld
def field_at(t: Any, path: Any):
    return field_at(t, path, Field(name="ROOT", type=t))


@ovld
def field_at(t: Any, path: str, f: Field):
    if not path:
        return f
    return recurse(t, path.lstrip(".").split("."), f)


@ovld(priority=10)
def field_at(t: Any, path: list, f: Field):
    if not path:
        return f
    else:
        return call_next(t, path, f)


@ovld
def field_at(t: type[dict], path: list, f: Field):
    (_, et) = get_args(t) or (str, object)
    _, *rest = path
    return recurse(et, rest, Field(name=f.name, type=et))


@ovld
def field_at(t: type[FieldModelizable], path: list, f: Field):
    m = model(t)
    curr, *rest = path
    for f2 in m.fields:
        if f2.serialized_name == curr:
            return recurse(f2.type, rest, f2)
    return None


@ovld(priority=1)
def field_at(t: type[ListModelizable], path: list, f: Field):
    m = model(t)
    curr, *rest = path
    try:
        int(curr)
    except ValueError:
        return None
    return recurse(m.element_field.type, rest, m.element_field)


@ovld
def field_at(t: type[UnionAlias], path: list, f: Field):
    for opt in get_args(t):
        if (rval := field_at(opt, path, f)) is not None:
            return rval
    return None


@ovld(priority=-1)
def field_at(t: Any, path: list, f: Field):
    return None



import importlib
import inspect
import sys
import typing
from types import GenericAlias, NoneType, UnionType
from typing import (
    Annotated,
    ForwardRef,
    Literal,
    TypeVar,
    Union,
    _GenericAlias,
    get_args,
    get_origin,
)

from ovld import class_check, dependent_check, parametrized_class_check

from .instructions import strip

try:
    from typing import TypeAliasType
except ImportError:  # pragma: no cover
    # It won't occur anyway
    class TypeAliasType:
        pass


@class_check
def UnionAlias(cls):
    return get_origin(cls) in (Union, UnionType)


def clsstring(cls):
    cls = getattr(cls, "original_type", cls)
    if args := typing.get_args(cls):
        origin = typing.get_origin(cls) or cls
        args = ", ".join(map(clsstring, args))
        return f"{origin.__name__}[{args}]"
    else:
        r = repr(cls)
        if r.startswith("<class "):
            return cls.__name__
        else:  # pragma: no cover
            return r


def basic_type(t):
    return get_origin(bt := strip(t)) or bt


#################
# evaluate_hint #
#################


class Indirect:
    def __init__(self, value):
        self.__value__ = value


def evaluate_hint(typ, ctx=None, lcl=None, typesub=None, seen=None):
    def get_eval_args():
        glb = ctx
        tsub = typesub
        local = lcl
        if glb is not None and not isinstance(glb, dict):
            if isinstance(glb, (GenericAlias, _GenericAlias)):
                origin = get_origin(glb)
                if hasattr(origin, "__type_params__"):
                    subs = {p: arg for p, arg in zip(origin.__type_params__, get_args(glb))}
                    tsub = {**subs, **(typesub or {})}
                glb = origin
            if hasattr(glb, "__type_params__"):
                local = {p.__name__: p for p in glb.__type_params__}
            glb = importlib.import_module(glb.__module__).__dict__
        return glb, local, tsub

    if isinstance(typ, str):
        if seen and (typ in seen):
            if seen[typ] is None:
                seen[typ] = Indirect(None)
            return seen[typ]
        else:
            glb, lcl, typesub = get_eval_args()
            if not seen:
                seen = {}
            seen[typ] = None
            rval = evaluate_hint(eval(typ, glb, lcl), glb, lcl, typesub, seen)
            if seen[typ] is not None:
                seen[typ].__value__ = rval
            return rval

    elif isinstance(typ, (UnionType, GenericAlias, _GenericAlias)):
        origin = get_origin(typ)
        if origin is Literal:
            return typ
        args = get_args(typ)
        if origin is Annotated:
            try:
                ot = evaluate_hint(args[0])
            except TypeError:
                ot = args[0]
            return Annotated[(ot, *args[1:])]
        if origin is UnionType:
            origin = Union
        new_args = [evaluate_hint(arg, ctx, lcl, typesub, seen) for arg in args]
        return origin[tuple(new_args)]

    elif isinstance(typ, TypeVar):
        return typesub.get(typ, typ) if typesub else typ

    elif isinstance(typ, ForwardRef):
        glb, lcl, _ = get_eval_args()
        if sys.version_info >= (3, 13):
            return typ._evaluate(glb, lcl, type_params=None, recursive_guard=frozenset())
        else:  # pragma: no cover
            return typ._evaluate(glb, lcl, recursive_guard=frozenset())

    elif isinstance(typ, type):
        return typ

    elif isinstance(typ, (Indirect, TypeAliasType)):
        return typ.__value__

    else:  # pragma: no cover
        raise TypeError("Cannot evaluate hint:", typ, type(typ))


####################
# Other type stuff #
####################


@parametrized_class_check
def JSONLike(t, bound=object):
    def _f(t, bound=object):
        origin = get_origin(t)
        if origin is typing.Union or origin is UnionType:
            return all(_f(t2) for t2 in get_args(t))
        if not isinstance(origin or t, type) or not issubclass(origin or t, bound):
            return False
        if t in (int, float, str, bool, NoneType):
            return True
        if origin is list:
            (et,) = get_args(t)
            return _f(et)
        if origin is dict:
            kt, vt = get_args(t)
            return (kt is str) and _f(vt)
        return False

    return _f(t, bound=bound)


@class_check
def IsLiteral(t):
    return get_origin(t) is Literal


@dependent_check
def JSON(obj: dict | list | str | int | float | None):
    if isinstance(obj, dict):
        return all(isinstance(k, str) and isinstance(v, JSON) for k, v in obj.items())
    elif isinstance(obj, list):
        return all(isinstance(v, JSON) for v in obj)
    else:
        return True


########
# Misc #
########


def check_signature(fn, flavor, expected):
    params = list(inspect.signature(fn).parameters.keys())
    nexpected = len(expected)
    expected = ", ".join(expected)
    if params[0] != "self" or len(params) != nexpected:  # pragma: no cover
        raise TypeError(
            f"{flavor} '{fn}' must define {nexpected} arguments: ({expected}). The first argument *must* be named 'self'."
        )


class MissingModule:  # pragma: no cover
    def __init__(self, feature, candidates):
        self.feature = feature
        self.candidates = candidates

    def __call__(self, *args, **kwargs):
        raise ImportError(
            f"{self.feature} requires ONE of the following packages to be installed:{self.candidates}"
        )


def import_any(feature, candidates):  # pragma: no cover
    proper_candidates = "".join(
        f"\n  - pip install {k.split(':')[0]}" for k, v in candidates.items() if v
    )
    candidates = {k.split(":")[-1]: v for k, v in candidates.items()}
    for modname, mapper in candidates.items():
        try:
            mod = importlib.import_module(modname)
            if mapper is None:
                result = MissingModule(feature, proper_candidates)
            else:
                result = mapper(mod)
            return result
        except ImportError:
            continue
    raise ImportError(
        f"{feature} requires ONE of the following packages to be installed:{proper_candidates}"
    )



import re
from datetime import date, datetime, timedelta
from pathlib import Path
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError

import pytest

from serieux import JSON, deserialize, serialize
from serieux.ctx import WorkingDirectory
from serieux.exc import ValidationError
from serieux.tell import tells

from .test_schema import schema

# Tests for non-principal types


#############
# Test date #
#############


def test_serialize_date():
    assert serialize(date, date(2023, 5, 15)) == "2023-05-15"
    assert serialize(list[date], [date(2023, 5, 15), date(2024, 1, 1)]) == [
        "2023-05-15",
        "2024-01-01",
    ]


def test_serialize_date_error():
    with pytest.raises(ValidationError, match="Cannot serialize"):
        serialize(date, "2023-05-15")
    with pytest.raises(ValidationError, match="Cannot serialize"):
        serialize(list[date], ["2023-05-15", "2024-01-01"])


def test_deserialize_date():
    assert deserialize(date, "2023-05-15") == date(2023, 5, 15)
    assert deserialize(list[date], ["2023-05-15", "2024-01-01"]) == [
        date(2023, 5, 15),
        date(2024, 1, 1),
    ]


def test_schema_date():
    assert schema(date) == {"type": "string", "format": "date"}


def test_tells_date():
    assert tells(date, str) == set()


#################
# Test datetime #
#################


def test_serialize_datetime():
    assert serialize(datetime, datetime(2023, 5, 15, 12, 30, 45)) == "2023-05-15T12:30:45"
    assert serialize(
        list[datetime], [datetime(2023, 5, 15, 12, 30, 45), datetime(2024, 1, 1, 0, 0, 0)]
    ) == ["2023-05-15T12:30:45", "2024-01-01T00:00:00"]


def test_deserialize_datetime():
    assert deserialize(datetime, "2023-05-15T12:30:45") == datetime(2023, 5, 15, 12, 30, 45)
    assert deserialize(list[datetime], ["2023-05-15T12:30:45", "2024-01-01T00:00:00"]) == [
        datetime(2023, 5, 15, 12, 30, 45),
        datetime(2024, 1, 1, 0, 0, 0),
    ]


def test_deserialize_datetime_timestamp():
    assert deserialize(datetime, 1746471553) == datetime.fromtimestamp(1746471553)


def test_schema_datetime():
    assert schema(datetime) == {"type": "string", "format": "date-time"}


def test_tells_datetime():
    assert tells(datetime, str) == set()


##################
# Test timedelta #
##################


def test_serialize_timedelta():
    assert serialize(timedelta, timedelta(seconds=42)) == "42s"
    assert serialize(timedelta, timedelta(seconds=42, microseconds=500000)) == "42500000us"
    assert serialize(timedelta, timedelta(seconds=-10)) == "-10s"
    assert serialize(list[timedelta], [timedelta(seconds=30), timedelta(days=5)]) == [
        "30s",
        "432000s",
    ]


def test_deserialize_timedelta():
    assert deserialize(timedelta, "42s") == timedelta(seconds=42)
    assert deserialize(timedelta, "42500000us") == timedelta(seconds=42, microseconds=500000)
    assert deserialize(timedelta, "-10s") == timedelta(seconds=-10)
    assert deserialize(timedelta, "6h30m") == timedelta(hours=6, minutes=30)
    assert deserialize(timedelta, "+8d") == timedelta(days=8)
    assert deserialize(timedelta, "4.5d") == timedelta(days=4, seconds=60 * 60 * 12)
    assert deserialize(timedelta, "1d12h") == timedelta(days=1, hours=12)
    assert deserialize(timedelta, "36h") == timedelta(days=1, hours=12)
    assert deserialize(list[timedelta], ["30s", "5d"]) == [
        timedelta(seconds=30),
        timedelta(days=5),
    ]
    with pytest.raises(ValidationError, match="is not a valid timedelta"):
        deserialize(timedelta, "1dd")
    with pytest.raises(ValidationError, match="is not a valid timedelta"):
        deserialize(timedelta, "1d3")
    with pytest.raises(ValidationError, match="is not a valid timedelta"):
        deserialize(timedelta, "1d3x")
    with pytest.raises(ValidationError, match="Could not convert"):
        deserialize(timedelta, "1.5.4d")


def test_schema_timedelta():
    assert schema(timedelta) == {
        "type": "string",
        "pattern": r"^[+-]?([\d.]+[dhms]|[\d.]+ms|[\d.]+us)+$",
    }


def test_tells_timedelta():
    assert tells(timedelta, str) == set()


#############
# Test Path #
#############


def test_serialize_path():
    assert serialize(Path, Path("hello/world.txt")) == "hello/world.txt"
    assert (
        serialize(Path, Path("hello/world.txt"), WorkingDirectory(directory=Path("hello")))
        == "world.txt"
    )


def test_deserialize_path():
    assert deserialize(Path, "hello/world.txt") == Path("hello/world.txt")
    assert deserialize(Path, "world.txt", WorkingDirectory(directory=Path("hello"))) == Path(
        "hello/world.txt"
    )


def test_schema_path():
    assert schema(Path) == {"type": "string"}


def test_tells_path():
    assert tells(Path, str) == set()


#################
# Test ZoneInfo #
#################


def test_serialize_zoneinfo():
    assert serialize(ZoneInfo, ZoneInfo("America/New_York")) == "America/New_York"
    assert serialize(ZoneInfo, ZoneInfo("UTC")) == "UTC"


def test_deserialize_zoneinfo():
    assert deserialize(ZoneInfo, "America/New_York") == ZoneInfo("America/New_York")
    assert deserialize(ZoneInfo, "UTC") == ZoneInfo("UTC")
    with pytest.raises(ZoneInfoNotFoundError, match="No time zone found"):
        deserialize(ZoneInfo, "Invalid/Timezone")


def test_schema_zoneinfo():
    assert schema(ZoneInfo) == {"type": "string"}


def test_tells_zoneinfo():
    assert tells(ZoneInfo, str) == set()


###################
# Test re.Pattern #
###################


def test_serialize_pattern():
    pattern = re.compile(r"\d+")
    assert serialize(re.Pattern, pattern) == r"\d+"

    pattern_with_flags = re.compile(r"hello", re.IGNORECASE)
    assert serialize(re.Pattern, pattern_with_flags) == "hello"


def test_deserialize_pattern():
    pattern = deserialize(re.Pattern, r"\d+")
    assert isinstance(pattern, re.Pattern)
    assert pattern.pattern == r"\d+"
    assert pattern.match("123") is not None
    assert pattern.match("abc") is None


def test_deserialize_complex_pattern():
    complex_pattern = deserialize(re.Pattern, r"^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$")
    assert complex_pattern.pattern == r"^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$"
    assert complex_pattern.match("test@example.com") is not None
    assert complex_pattern.match("invalid-email") is None


def test_deserialize_pattern_invalid():
    with pytest.raises(Exception, match="bad character range"):
        deserialize(re.Pattern, r"[z-a]")  # Invalid character range


def test_serialize_deserialize_dict_pattern_str():
    pattern1 = re.compile(r"\d+")
    pattern2 = re.compile(r"hello")
    d = {pattern1: "numbers", pattern2: "greeting"}

    serialized = serialize(dict[re.Pattern, str], d)
    assert serialized == {r"\d+": "numbers", "hello": "greeting"}

    deserialized = deserialize(
        dict[re.Pattern, str],
        {r"\d+": "numbers", "hello": "greeting"},
    )
    assert deserialized == d


def test_schema_pattern():
    assert schema(re.Pattern) == {"type": "string"}


def test_tells_pattern():
    assert tells(re.Pattern, str) == set()


#############
# Test JSON #
#############


class NotJson:
    pass


def test_serialize_json():
    data = {"a": 1, "b": [2, 3], "c": {"d": "hello"}}
    assert serialize(JSON, data) == data
    assert serialize(list[JSON], [data, 42, "foo"]) == [data, 42, "foo"]
    with pytest.raises(ValidationError):
        serialize(JSON, NotJson())


def test_deserialize_json():
    data = {"a": 1, "b": [2, 3], "c": {"d": "hello"}}
    assert deserialize(JSON, data) == data
    assert deserialize(list[JSON], [data, 42, "foo"]) == [data, 42, "foo"]
    with pytest.raises(ValidationError):
        deserialize(JSON, NotJson())


def test_schema_json():
    assert schema(JSON) == {}



import re
from dataclasses import dataclass, field
from functools import partial

from ovld import Medley, call_next, ovld
from ovld.dependent import Regexp
from rich.pretty import pprint

from serieux import LazyProxy, Serieux, deserialize
from serieux.ctx import Trail

##################
# Implementation #
##################


class Variables(Trail):
    refs: dict[tuple[str, ...], object] = field(default_factory=dict)


def evaluate(expr, ctx):
    current = ctx.trail[:-1]
    while True:
        lcl = vars(ctx.refs[current])
        try:
            return eval(expr, lcl, lcl)
        except NameError:
            pass
        if not current:
            raise Exception(f"Failed to evaluate expression: {expr}")
        current = current[:-1]


@Serieux.extend
class VarInterpolation(Medley):
    @ovld(priority=3)
    def deserialize(self, typ: type[object], value: object, ctx: Variables):
        rval = call_next(typ, value, ctx)
        ctx.refs[ctx.trail] = rval
        return rval

    @ovld(priority=2)
    def deserialize(self, typ: type[object], value: Regexp[r"^\$\{.+\}$"], ctx: Variables):
        expr = value.lstrip("${").rstrip("}")
        return LazyProxy(partial(evaluate, expr, ctx))

    @ovld(priority=1)
    def deserialize(self, typ: type[object], value: Regexp[r"\$\{.+\}"], ctx: Variables):
        def interpolate():
            def repl(match):
                return str(evaluate(match.group(1), ctx))

            return re.sub(r"\$\{(.+)\}", repl, value)

        return LazyProxy(interpolate)


#################
# Demonstration #
#################


@dataclass
class Philosopher:
    name: str
    school_of_thought: str
    birth_year: int


@dataclass
class DebatingSociety:
    founder: Philosopher
    rival_philosopher: Philosopher
    meeting_place: str
    year_established: int


def main():
    plato = {"name": "Plato", "school_of_thought": "Platonism", "birth_year": -428}

    aristotle = {
        "name": "Aristotle",
        "school_of_thought": "Team ${name}",
        "birth_year": -384,
    }

    academy_serialized = {
        "founder": plato,
        "rival_philosopher": aristotle,
        "meeting_place": "Athens",
        "year_established": "${founder.birth_year}",
    }
    print("\n== Serialized ==\n")
    pprint(academy_serialized)

    academy = deserialize(DebatingSociety, academy_serialized, Variables())
    print("\n== Deserialized ==\n")
    pprint(academy)
    assert academy.year_established == -428
    assert academy.rival_philosopher.school_of_thought == "Team Aristotle"


if __name__ == "__main__":
    main()



from dataclasses import dataclass

from serieux import deserialize
from serieux.ctx import Patcher
from serieux.features.prompt import Promptable


@dataclass
class Person:
    # NAME of the person
    name: str
    # AGE of the person
    age: int
    # MADNESS of the person
    mad: bool


def _prompter(prompts):
    def resolve(ctx, prompt):
        assert prompt in prompts
        return prompts[prompt]

    return resolve


def test_resolve_prompt():
    value = deserialize(
        int,
        "${prompt:Enter your age}",
        Promptable(prompt_function=_prompter({"Enter your age": "42"})),
    )
    assert value == 42


def test_resolve_prompt_boolean():
    value = deserialize(
        bool,
        "${prompt:Are you sure?}",
        Promptable(prompt_function=_prompter({"Are you sure?": "yes"})),
    )
    assert value is True


def test_resolve_prompt_string():
    value = deserialize(
        str,
        "${prompt:What is your name?}",
        Promptable(prompt_function=_prompter({"What is your name?": "John Doe"})),
    )
    assert value == "John Doe"


TEST_YAML = """
name: "${prompt:}"
age: "${prompt:}"
mad: "${prompt:Is the person mad?}"
"""

MODIFIED_YAML = """
name: "John Doe"
age: 42
mad: true
"""


def test_prompt_with_patcher(tmp_path):
    # Create a YAML file with a prompt directive
    yaml_file = tmp_path / "test.yaml"
    yaml_file.write_text(TEST_YAML)

    # Create a prompter that returns fixed values
    prompter = _prompter(
        {"NAME of the person": "John Doe", "AGE of the person": "42", "Is the person mad?": "yes"}
    )

    # Deserialize with Patcher
    ctx = Promptable(prompt_function=prompter) + Patcher()
    result = deserialize(Person, yaml_file, ctx)

    # Verify the deserialized result
    assert result == Person(name="John Doe", age=42, mad=True)

    # Apply patches
    ctx.apply_patches()

    # Verify the file was modified
    modified_content = yaml_file.read_text()
    assert modified_content == MODIFIED_YAML



import importlib.metadata
from dataclasses import dataclass
from pathlib import Path

from .abc import FileFormat


@dataclass
class FileSource:
    path: Path
    format: FileFormat = None
    field: str = None

    def __post_init__(self):
        if not isinstance(self.format, FileFormat):
            self.format = find(self.path, suffix=self.format)

    @classmethod
    def serieux_from_string(cls, incl):
        pth, at, fld = incl.partition("@")
        return cls(Path(pth.strip()), field=fld if at else None)


class FormatRegistry(dict):
    def __missing__(self, item):
        match importlib.metadata.entry_points(group="serieux.formats", name=item):
            case [ep, *__]:
                ff_cls = ep.load()
                ff = ff_cls()
                self[item] = ff
                return ff
            case _:
                raise ImportError(f"Format `{item}` is not recognized.")


registry = FormatRegistry()


def find(p: Path, suffix: str | None = None):
    if suffix is None:
        suffix = p.suffix
    suffix = suffix.lstrip(".")
    return registry[suffix]


def load(p: Path, suffix: str | None = None):
    return find(p, suffix).load(p)


def dump(p: Path, data: object, suffix: str | None = None):
    return find(p, suffix).dump(p, data)


def loads(s: str, suffix: str | None = None):
    return find(None, suffix).loads(s)


def dumps(data: object, suffix: str | None = None):
    return find(None, suffix).dumps(data)



from typing import Any, Callable, Literal

from ovld import ovld

from ..ctx import Patcher
from ..model import field_at
from .interpol import Environment, decode_string


def default_prompt(ctx, prompt):  # pragma: no cover
    return input(
        f"\033[1;36m[{'.'.join(str(x) for x in ctx.trail)}]\033[0m \033[1;33m{prompt}\033[0m\n\033[1;32m>\033[0m "
    )


class Promptable(Environment):
    prompt_function: Callable[[str], str] = default_prompt

    @ovld
    def resolve_variable(self, t: Any, method: Literal["prompt"], expr: str, /):
        if not expr:
            objt, _, field = self.full_trail[-1]
            fld = field_at(objt, [field])
            expr = (fld and fld.description) or "Enter value"
        value = decode_string(t, self.prompt_function(self, expr))
        if isinstance(self, Patcher):
            self.declare_patch(value)
        return value



from itertools import pairwise

from serieux.priority import (
    DEFAULT,
    HI1,
    HI2,
    HI3,
    HI4,
    HI5,
    HI6,
    HI7,
    HI8,
    HI9,
    HIGH,
    LO1,
    LO2,
    LO3,
    LO4,
    LO5,
    LOW,
    MAX,
    MIN,
    STD,
    STD1,
    STD2,
    STD3,
    STD4,
    STD5,
    NamedInteger,
    PriorityLevel,
)


def test_ni_creation():
    ni = NamedInteger(42, "test")
    assert ni == 42
    assert int(ni) == 42
    assert ni._name == "test"
    assert isinstance(ni, int)


def test_ni_string_representation():
    ni = NamedInteger(100, "MAX")
    assert str(ni) == "MAX"
    assert repr(ni) == "MAX"


def test_ni_arithmetic_operations():
    ni = NamedInteger(10, "TEN")
    assert ni + 5 == 15
    assert ni - 3 == 7
    assert ni * 2 == 20
    assert ni // 2 == 5


def test_ni_comparison_operations():
    ni1 = NamedInteger(5, "FIVE")
    ni2 = NamedInteger(10, "TEN")
    assert ni1 < ni2
    assert ni2 > ni1


def test_pl_next_method():
    ni = NamedInteger(5, "FIVE")
    pl = PriorityLevel(ni)
    assert pl.next() == (5, 1)
    assert pl.next() == (5, 2)
    assert pl.next() == (5, 3)
    assert pl.next().next() == (5, 4, 1)


def test_call_method():
    ni = NamedInteger(5, "FIVE")
    pl = PriorityLevel(ni)
    assert pl(10, "extra") == (ni, 10, "extra")


def test_order():
    seq = [
        MAX,
        HI9,
        HI8,
        HI7,
        HI6,
        HI5,
        HI4,
        HI3,
        HI2,
        HI1,
        DEFAULT,
        STD5,
        STD4,
        STD3,
        STD2,
        STD1,
        LO1,
        LO2,
        LO3,
        LO4,
        LO5,
        MIN,
    ]
    assert all(a > b for a, b in pairwise(seq))


def test_aliases():
    assert DEFAULT == (0,)
    assert STD == STD1
    assert HIGH == HI1
    assert LOW == LO1



from dataclasses import replace
from pathlib import Path
from typing import Any

from ovld import call_next, ovld, recurse
from ovld.dependent import HasKey

from ..ctx import Context, Sourced, WorkingDirectory
from ..exc import ValidationError
from ..formats import FileSource
from ..priority import HI1, MIN
from ..utils import clsstring
from .partial import PartialBuilding, Sources

include_field = "$include"


class FromFile(PartialBuilding):
    def deserialize(self, t: Any, obj: FileSource, ctx: Context):
        pth = obj.path
        if isinstance(ctx, WorkingDirectory):
            pth = ctx.directory / pth.expanduser()
        if not pth.exists():
            raise ValidationError(f"File '{pth.absolute()}' does not exist", ctx=ctx)
        data = obj.format.load(pth)
        if obj.field:
            for f in obj.field.split("."):
                data = data[f]
        ctx = ctx + Sourced(
            origin=pth,
            directory=pth.parent.absolute(),
            format=obj.format,
            source_trail=getattr(ctx, "trail", ()),
        )
        return recurse(t, data, ctx)

    def deserialize(self, t: Any, obj: Path, ctx: Context):
        return recurse(t, FileSource(obj), ctx)


class IncludeFile(FromFile):
    @ovld(priority=HI1)
    def deserialize(self, t: Any, obj: HasKey[include_field], ctx: Context):
        obj = dict(obj)
        paths = recurse(FileSource | list[FileSource], obj.pop(include_field), ctx)
        match paths:
            case [pth] | (FileSource() as pth):
                if obj:
                    return recurse(t, Sources(pth, obj), ctx)
                else:
                    return recurse(t, pth, ctx)
            case _:
                if obj:  # pragma: no cover
                    return recurse(t, Sources(*paths, obj), ctx)
                else:
                    return recurse(t, Sources(*paths), ctx)

    @ovld(priority=MIN)
    def deserialize(self, t: Any, obj: str, ctx: WorkingDirectory):
        if "." not in obj or obj.rsplit(".", 1)[1].isnumeric():
            return call_next(t, obj, ctx)

        src = recurse(FileSource, obj, ctx)
        src = replace(src, path=ctx.directory / src.path)

        if src.path.exists():
            return recurse(t, src, ctx)
        else:
            raise ValidationError(
                f"Tried to read {obj!r} as a configuration file (at path '{src.path}')"
                f" to deserialize into object of type {clsstring(t)},"
                " but there was no such file."
            )



import importlib
from pathlib import Path

import pytest

examples = Path(__file__).parent / "../examples"


example_files = [f for f in examples.glob("**/*.py") if not f.name.startswith("_")]


@pytest.mark.parametrize("file", example_files, ids=lambda f: f.stem)
def test_example(file, file_regression, capsys, fresh_serieux):
    module_name = file.stem
    spec = importlib.util.spec_from_file_location(module_name, file)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    module.main()
    if not getattr(module.main, "do_not_test_output", False):
        captured = capsys.readouterr()
        file_regression.check(captured.out)



from dataclasses import dataclass
from pathlib import Path
from types import NoneType
from typing import Annotated, Any

from ovld import Code, ovld, recurse

from .instructions import pushdown
from .model import FieldModelizable, ListModelizable, StringModelizable, model


class Tell:
    def __lt__(self, other):
        return self.cost() < other.cost()

    def cost(self):  # pragma: no cover
        return 1


@dataclass(frozen=True)
class KeyTell(Tell):
    key: str

    def gen(self, arg):
        return Code("(isinstance($arg, dict) and $k in $arg)", arg=arg, k=self.key)

    def cost(self):
        return 2


@dataclass(frozen=True)
class KeyValueTell(Tell):
    key: str
    value: object

    def gen(self, arg):
        return Code(
            "(isinstance($arg, dict) and $k in $arg and $arg[$k] == $v)",
            arg=arg,
            k=self.key,
            v=self.value,
        )

    def cost(self):  # pragma: no cover
        return 3


@ovld
def tells(expected: type[int], given: type[int]):
    return set()


@ovld
def tells(expected: type[str] | type[Path] | type[StringModelizable], given: type[str]):
    return set()


@ovld
def tells(expected: type[bool], given: type[bool]):
    return set()


@ovld
def tells(expected: type[float], given: type[float]):
    return set()


@ovld
def tells(expected: type[NoneType], given: type[NoneType]):
    return set()


@ovld
def tells(expected: type[dict], given: type[dict]):
    return set()


@ovld
def tells(expected: type[FieldModelizable], given: type[dict]):
    m = model(expected)
    return {KeyTell(f.serialized_name) for f in m.fields}


@ovld
def tells(expected: type[ListModelizable], given: type[list]):
    return set()


@ovld
def tells(expected: type[Annotated], given: Any):
    return recurse(pushdown(expected), given)


@ovld(priority=-1)
def tells(expected: Any, given: Any):
    return None
